# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import warnings
from functools import partial

import numpy as np
import pytest

from power_grid_model._core.data_handling import create_output_data
from power_grid_model._core.dataset_definitions import ComponentType as CT, DatasetType as DT
from power_grid_model._core.power_grid_core import VoidPtr
from power_grid_model._core.power_grid_dataset import CMutableDataset
from power_grid_model._core.power_grid_meta import initialize_array, power_grid_meta_data
from power_grid_model._core.utils import is_columnar
from power_grid_model.enum import ComponentAttributeFilterOptions


def columnar_array(component_type, n_components, attributes=None, batch_size_tuple=()):
    component_dtype = power_grid_meta_data[DT.sym_output][component_type].dtype
    required_attributes = (
        set(component_dtype.names) & set(attributes) if attributes is not None else component_dtype.names
    )
    return {
        attr: np.empty((n_components, *batch_size_tuple), dtype=component_dtype[attr]) for attr in required_attributes
    }


def row_array(component_type, n_components, batch_size_tuple=()):
    return initialize_array(DT.sym_output, component_type, (n_components, *batch_size_tuple))


@pytest.fixture(params=[1, 15])
def batch_size(request):
    return request.param


@pytest.mark.parametrize(
    ("output_component_types", "expected_fns"),
    [
        (
            None,
            {
                CT.node: partial(row_array, component_type=CT.node, n_components=4),
                CT.sym_load: partial(row_array, component_type=CT.sym_load, n_components=3),
                CT.source: partial(row_array, component_type=CT.source, n_components=1),
            },
        ),
        (
            [CT.node, CT.sym_load],
            {
                CT.node: partial(row_array, component_type=CT.node, n_components=4),
                CT.sym_load: partial(row_array, component_type=CT.sym_load, n_components=3),
            },
        ),
        (
            {CT.node, CT.sym_load},
            {
                CT.node: partial(row_array, component_type=CT.node, n_components=4),
                CT.sym_load: partial(row_array, component_type=CT.sym_load, n_components=3),
            },
        ),
        pytest.param(
            ComponentAttributeFilterOptions.relevant,
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=None),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=None),
                CT.source: partial(columnar_array, component_type=CT.source, n_components=1, attributes=None),
            },
        ),
        pytest.param(
            {
                CT.node: ComponentAttributeFilterOptions.everything,
                CT.sym_load: ComponentAttributeFilterOptions.relevant,
            },
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=None),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=None),
            },
        ),
        pytest.param(
            {CT.node: ComponentAttributeFilterOptions.relevant, CT.sym_load: ["p"]},
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=None),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=["p"]),
            },
        ),
        pytest.param(
            {CT.node: None, CT.sym_load: ["p"]},
            {
                CT.node: partial(row_array, component_type=CT.node, n_components=4),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=["p"]),
            },
        ),
        pytest.param(
            {CT.node: [], CT.sym_load: []},
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=[]),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=[]),
            },
        ),
        pytest.param(
            {CT.node: [], CT.sym_load: ["p"]},
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=[]),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=["p"]),
            },
        ),
        pytest.param(
            {CT.node: ["u"], CT.sym_load: ["p"]},
            {
                CT.node: partial(columnar_array, component_type=CT.node, n_components=4, attributes=["u"]),
                CT.sym_load: partial(columnar_array, component_type=CT.sym_load, n_components=3, attributes=["p"]),
            },
        ),
    ],
)
def test_create_output_data(output_component_types, expected_fns, batch_size):
    """Test output_data creation. Batch size is set later in the test."""
    all_component_count = {CT.node: 4, CT.sym_load: 3, CT.source: 1}
    actual = create_output_data(
        output_component_types=output_component_types,
        output_type=DT.sym_output,
        all_component_count=all_component_count,
        is_batch=batch_size != 1,
        batch_size=batch_size,
    )

    expected = {comp: fn(batch_size_tuple=(batch_size,)) for comp, fn in expected_fns.items()}
    assert actual.keys() == expected.keys()
    for comp, fn in expected.items():
        if not is_columnar(fn):
            assert actual[comp].dtype == fn.dtype
        elif fn == dict():
            # Empty attributes columnar
            assert actual[comp] == fn
        else:
            assert actual[comp].keys() == fn.keys()
            assert all(actual[comp][attr].dtype == fn[attr].dtype for attr in fn)


def test_dtype_compatibility_check_normal():
    nodes = initialize_array(DT.sym_output, CT.node, (1, 2))
    nodes_ptr = nodes.ctypes.data_as(VoidPtr)

    data = {CT.node: nodes}
    mutable_dataset = CMutableDataset(data, DT.sym_output)
    buffer_views = mutable_dataset.get_buffer_views()

    assert buffer_views[0].data.value == nodes_ptr.value


def test_dtype_compatibility_check_compatible():
    nodes = initialize_array(DT.sym_output, CT.node, 4)
    nodes = nodes[::2]
    nodes_ptr = nodes.ctypes.data_as(VoidPtr)

    data = {CT.node: nodes}
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        mutable_dataset = CMutableDataset(data, DT.sym_output)
        buffer_views = mutable_dataset.get_buffer_views()

    assert buffer_views[0].data.value != nodes_ptr.value


def test_dtype_compatibility_check__error():
    nodes = initialize_array(DT.sym_output, CT.node, (1, 2))
    data = {CT.node: nodes.astype(nodes.dtype.newbyteorder("S"))}
    with pytest.raises(ValueError, match="Data type does not match schema"):
        CMutableDataset(data, DT.sym_output)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Configuration file for the Sphinx documentation builder.
"""

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "power-grid-model"
copyright = "Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>"
author = "Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>"

# -- Setup

import os  # noqa: E402

# Fix linking in github and rtd
link_head_gh = "https://github.com/PowerGridModel/power-grid-model/"
if "READTHEDOCS" in os.environ:
    import git

    commit_version = git.Repo(search_parent_directories=True).head.object.hexsha
    link_head_gh_blob = link_head_gh + "blob/" + commit_version
    link_head_gh_tree = link_head_gh + "tree/" + commit_version
    pgm_project_root = ""
    pgm_project_contribution = pgm_project_root + "/contribution"
else:
    link_head_gh_blob = ""
    link_head_gh_tree = ""
    pgm_project_root = "https://github.com/PowerGridModel/.github/blob/main"
    pgm_project_contribution = pgm_project_root


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "breathe",
    "sphinx.ext.napoleon",
    "hoverxref.extension",
    "myst_nb",
    "sphinxcontrib.mermaid",
    "sphinxcontrib.tikz",
]

templates_path = ["_templates"]

tikz_latex_preamble = "\\usepackage{circuitikz}"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

exclude_patterns = [
    "*/.ipynb_checkpoints/*",
    "_build/**/*",
]

# -- myst parser and myst_nb config ------------------------------------------------------
# label references for depth of headers: label name in anchor slug structure
myst_heading_anchors = 4
# execute jupter notebooks output before building webpage
nb_execution_mode = "off"
nb_execution_excludepatterns = ["*/_build/*"]

# Extentions in myst
myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "substitution",
]
# Global substitutions
myst_substitutions = {
    "gh_link_head_blob": link_head_gh_blob,
    "gh_link_head_tree": link_head_gh_tree,
    "pgm_project_root": pgm_project_root,
    "pgm_project_contribution": pgm_project_contribution,
}


# -- hoverxref config --------------------------------------------------------
# hover tooltip on python classes
# hoverx links are in sphinx style: {ref} `docs_path\file:Exact Section Name`
hoverxref_domains = [
    "py",
]
hoverxref_default_type = "tooltip"
hoverxref_role_types = {
    "hoverxref": "tooltip",
    "ref": "modal",  # for hoverxref_auto_ref config
    "confval": "tooltip",  # for custom object
    "mod": "tooltip",  # for Python Sphinx Domain
    "class": "tooltip",  # for Python Sphinx Domain
    "obj": "tooltip",
    "function": "tooltip",
}

# -- sphinx.autodoc config ---------------------------------------------------
autodoc_default_options = {
    "members": None,
    "member-order": "bysource",
    "special-members": "__init__",
    "undoc-members": False,
    "exclude-members": "__weakref__",
}

# -- sphinx.autosectionlabel config -------------------------------------------
autosectionlabel_prefix_document = True

# -- Add google html -----------------------------------
html_extra_path = ["google6d726d2d56f95e32.html"]

# config doxygen for C API
breathe_projects = {"power_grid_model_c": "./doxygen/build/xml/"}
breathe_default_project = "power_grid_model_c"

# Override theme CSS with style adjustments of our own
html_style = "css/custom.css"



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from copy import copy

import pytest

from power_grid_model import PowerGridModel
from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.power_grid_meta import initialize_array
from power_grid_model.enum import (
    AngleMeasurementType,
    CalculationMethod,
    LoadGenType,
    MeasuredTerminalType,
    TapChangingStrategy,
)
from power_grid_model.errors import (
    AutomaticTapInputError,
    ConflictID,
    ConflictingAngleMeasurementType,
    ConflictVoltage,
    IDNotFound,
    IDWrongType,
    InvalidBranch,
    InvalidBranch3,
    InvalidCalculationMethod,
    InvalidMeasuredObject,
    InvalidRegulatedObject,
    InvalidTransformerClock,
    IterationDiverge,
    MissingCaseForEnumError,
    NotObservableError,
)

from .utils import PowerGridModelWithExt


def test_empty_model():
    model = PowerGridModel.__new__(PowerGridModel)
    with pytest.raises(TypeError):
        model.calculate_power_flow()
    with pytest.raises(TypeError):
        model.copy()
    with pytest.raises(TypeError):
        _n = model.all_component_count
    with pytest.raises(TypeError):
        copy(model)


def test_unknown_component_types():
    model = PowerGridModel(input_data={})
    with pytest.raises(KeyError, match=r"fake_type"):
        model.calculate_power_flow(output_component_types={"fake_type"})


def test_handle_conflict_voltage_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [10.0e3, 20.0e3]

    line_input = initialize_array(DatasetType.input, ComponentType.line, 1)
    line_input["id"] = [2]
    line_input["from_node"] = [0]
    line_input["to_node"] = [1]

    with pytest.raises(ConflictVoltage):
        PowerGridModel(input_data={ComponentType.node: node_input, ComponentType.line: line_input})


def test_handle_missing_case_for_enum_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [1e4, 4e2]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [2]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [10.0e3]

    transformer_input = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    transformer_input["id"] = [3]
    transformer_input["from_node"] = [0]
    transformer_input["to_node"] = [1]
    transformer_input["from_status"] = [1]
    transformer_input["to_status"] = [1]
    transformer_input["winding_from"] = [2]
    transformer_input["winding_to"] = [1]
    transformer_input["clock"] = [5]
    transformer_input["tap_side"] = [0]

    transformer_tap_regulator_input = initialize_array(DatasetType.input, ComponentType.transformer_tap_regulator, 1)
    transformer_tap_regulator_input["id"] = [4]
    transformer_tap_regulator_input["regulated_object"] = [3]
    transformer_tap_regulator_input["status"] = [1]
    transformer_tap_regulator_input["control_side"] = [127]

    with pytest.raises(MissingCaseForEnumError):
        PowerGridModelWithExt(
            input_data={
                ComponentType.node: node_input,
                ComponentType.transformer: transformer_input,
                ComponentType.source: source_input,
                ComponentType.transformer_tap_regulator: transformer_tap_regulator_input,
            }
        )


def test_handle_invalid_branch_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [0.0]

    line_input = initialize_array(DatasetType.input, ComponentType.line, 1)
    line_input["id"] = [1]
    line_input["from_node"] = [0]
    line_input["to_node"] = [0]

    with pytest.raises(InvalidBranch):
        PowerGridModel(input_data={ComponentType.node: node_input, ComponentType.line: line_input})


def test_handle_invalid_branch3_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [0.0]

    three_winding_transformer_input = initialize_array(DatasetType.input, ComponentType.three_winding_transformer, 1)
    three_winding_transformer_input["id"] = [1]
    three_winding_transformer_input["node_1"] = [0]
    three_winding_transformer_input["node_2"] = [0]
    three_winding_transformer_input["node_3"] = [0]

    with pytest.raises(InvalidBranch3):
        PowerGridModel(
            input_data={
                ComponentType.node: node_input,
                ComponentType.three_winding_transformer: three_winding_transformer_input,
            }
        )


def test_handle_invalid_transformer_clock_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [0.0, 0.0]

    transformer_input = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    transformer_input["id"] = [2]
    transformer_input["from_node"] = [0]
    transformer_input["to_node"] = [1]
    transformer_input["clock"] = [-1]

    with pytest.raises(InvalidTransformerClock):
        PowerGridModel(input_data={ComponentType.node: node_input, ComponentType.transformer: transformer_input})


@pytest.mark.skip(reason="TODO")
def test_handle_sparse_matrix_error():
    pass


def test_handle_not_observable_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [10.0e3]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [1]
    source_input["node"] = [0]
    source_input["status"] = [1]

    sym_load_input = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
    sym_load_input["id"] = [2]
    sym_load_input["node"] = [0]
    sym_load_input["status"] = [1]
    sym_load_input["type"] = [LoadGenType.const_power]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.source: source_input,
            ComponentType.sym_load: sym_load_input,
        }
    )
    with pytest.raises(NotObservableError):
        model.calculate_state_estimation(calculation_method=CalculationMethod.iterative_linear)


def test_handle_iteration_diverge_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [100.0]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [1]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [1.0]
    source_input["sk"] = [1000.0]
    source_input["rx_ratio"] = [0.0]

    sym_load_input = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
    sym_load_input["id"] = [2]
    sym_load_input["node"] = [0]
    sym_load_input["status"] = [1]
    sym_load_input["type"] = [LoadGenType.const_current]
    sym_load_input["p_specified"] = [0.0]
    sym_load_input["q_specified"] = [500.0]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.source: source_input,
            ComponentType.sym_load: sym_load_input,
        }
    )
    with pytest.raises(IterationDiverge):
        model.calculate_power_flow(max_iterations=1, error_tolerance=1.0e-100)


def test_handle_conflict_id_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 0]
    node_input["u_rated"] = [0.0, 0.0]

    with pytest.raises(ConflictID):
        PowerGridModel(input_data={ComponentType.node: node_input})


def test_handle_id_not_found_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [0.0]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [1]
    source_input["node"] = [99]
    source_input["status"] = [1]
    source_input["u_ref"] = [0.0]

    with pytest.raises(IDNotFound):
        PowerGridModel(input_data={ComponentType.node: node_input, ComponentType.source: source_input})


@pytest.mark.parametrize("sensor_type", [ComponentType.sym_power_sensor, ComponentType.sym_current_sensor])
def test_handle_invalid_measured_object_error(sensor_type):
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [0.0, 0.0]

    link_input = initialize_array(DatasetType.input, ComponentType.link, 1)
    link_input["id"] = [2]
    link_input["from_node"] = [0]
    link_input["to_node"] = [1]

    sensor_input = initialize_array(DatasetType.input, sensor_type, 1)
    sensor_input["id"] = [3]
    sensor_input["measured_object"] = [2]
    sensor_input["measured_terminal_type"] = [MeasuredTerminalType.branch_from]

    with pytest.raises(InvalidMeasuredObject):
        PowerGridModel(
            input_data={ComponentType.node: node_input, ComponentType.link: link_input, sensor_type: sensor_input}
        )


def test_handle_invalid_regulated_object_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [1e4, 4e2]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [2]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [10.0e3]

    transformer_tap_regulator_input = initialize_array(DatasetType.input, ComponentType.transformer_tap_regulator, 1)
    transformer_tap_regulator_input["id"] = [3]
    transformer_tap_regulator_input["regulated_object"] = [2]

    with pytest.raises(InvalidRegulatedObject):
        PowerGridModel(
            input_data={
                ComponentType.node: node_input,
                ComponentType.source: source_input,
                ComponentType.transformer_tap_regulator: transformer_tap_regulator_input,
            }
        )


def test_handle_id_wrong_type_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [0.0]

    sym_power_sensor_input = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 1)
    sym_power_sensor_input["id"] = [1]
    sym_power_sensor_input["measured_object"] = [0]
    sym_power_sensor_input["measured_terminal_type"] = [MeasuredTerminalType.branch_from]

    with pytest.raises(IDWrongType):
        PowerGridModel(
            input_data={ComponentType.node: node_input, ComponentType.sym_power_sensor: sym_power_sensor_input}
        )


def test_handle_invalid_calculation_method_error():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 1)
    node_input["id"] = [0]
    node_input["u_rated"] = [10.0e3]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [1]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [10.0e3]

    sym_load_input = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
    sym_load_input["id"] = [2]
    sym_load_input["node"] = [0]
    sym_load_input["status"] = [1]
    sym_load_input["type"] = [LoadGenType.const_power]
    sym_load_input["p_specified"] = [10.0e3]
    sym_load_input["q_specified"] = [1.0e3]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.source: source_input,
            ComponentType.sym_load: sym_load_input,
        }
    )
    with pytest.raises(InvalidCalculationMethod):
        model.calculate_power_flow(calculation_method=CalculationMethod.iec60909)


def test_transformer_tap_regulator_control_side_not_closer_to_source():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [150e3, 10e3]

    transformer_input = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    transformer_input["id"] = [2]
    transformer_input["from_node"] = [0]
    transformer_input["to_node"] = [1]
    transformer_input["from_status"] = [1]
    transformer_input["to_status"] = [1]
    transformer_input["u1"] = [1e4]
    transformer_input["u2"] = [4e2]
    transformer_input["sn"] = [1e5]
    transformer_input["uk"] = [0.1]
    transformer_input["pk"] = [1e3]
    transformer_input["i0"] = [1.0e-6]
    transformer_input["p0"] = [0.1]
    transformer_input["winding_from"] = [1]
    transformer_input["winding_to"] = [1]
    transformer_input["clock"] = [0]
    transformer_input["tap_side"] = [1]
    transformer_input["tap_pos"] = [0]
    transformer_input["tap_nom"] = [0]
    transformer_input["tap_min"] = [-1]
    transformer_input["tap_max"] = [1]
    transformer_input["tap_size"] = [100]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [3]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [1.0]

    transformer_tap_regulator_input = initialize_array(DatasetType.input, ComponentType.transformer_tap_regulator, 1)
    transformer_tap_regulator_input["id"] = [4]
    transformer_tap_regulator_input["regulated_object"] = [2]
    transformer_tap_regulator_input["status"] = [1]
    transformer_tap_regulator_input["u_set"] = [10]
    transformer_tap_regulator_input["u_band"] = [200]
    transformer_tap_regulator_input["control_side"] = [0]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.transformer: transformer_input,
            ComponentType.source: source_input,
            ComponentType.transformer_tap_regulator: transformer_tap_regulator_input,
        }
    )
    with pytest.raises(AutomaticTapInputError):
        model.calculate_power_flow(tap_changing_strategy=TapChangingStrategy.min_voltage_tap, decode_error=True)


def test_automatic_tap_changing():
    model = PowerGridModel(input_data={})
    model.calculate_power_flow(tap_changing_strategy=TapChangingStrategy.min_voltage_tap)


def test_conflicting_angle_measurement_type() -> None:
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [10e3, 10e3]

    line_input = initialize_array(DatasetType.input, ComponentType.line, 1)
    line_input["id"] = [2]
    line_input["from_node"] = [0]
    line_input["to_node"] = [1]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [3]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [1.0]

    sym_voltage_sensor_input = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 1)
    sym_voltage_sensor_input["id"] = [4]
    sym_voltage_sensor_input["measured_object"] = [0]
    sym_voltage_sensor_input["u_sigma"] = [1.0]
    sym_voltage_sensor_input["u_measured"] = [1.0]
    sym_voltage_sensor_input["u_angle_measured"] = [0.0]

    sym_current_sensor_input = initialize_array(DatasetType.input, ComponentType.sym_current_sensor, 2)
    sym_current_sensor_input["id"] = [5, 6]
    sym_current_sensor_input["measured_object"] = [2, 2]
    sym_current_sensor_input["measured_terminal_type"] = [
        MeasuredTerminalType.branch_from,
        MeasuredTerminalType.branch_from,
    ]
    sym_current_sensor_input["angle_measurement_type"] = [
        AngleMeasurementType.local_angle,
        AngleMeasurementType.global_angle,
    ]
    sym_current_sensor_input["i_sigma"] = [1.0, 1.0]
    sym_current_sensor_input["i_angle_sigma"] = [1.0, 1.0]
    sym_current_sensor_input["i_measured"] = [0.0, 0.0]
    sym_current_sensor_input["i_angle_measured"] = [0.0, 0.0]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.line: line_input,
            ComponentType.source: source_input,
            ComponentType.sym_voltage_sensor: sym_voltage_sensor_input,
            ComponentType.sym_current_sensor: sym_current_sensor_input,
        }
    )

    with pytest.raises(ConflictingAngleMeasurementType):
        model._calculate_state_estimation(decode_error=True)


def test_global_current_measurement_without_voltage_angle() -> None:
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [0, 1]
    node_input["u_rated"] = [10e3, 10e3]

    line_input = initialize_array(DatasetType.input, ComponentType.line, 1)
    line_input["id"] = [2]
    line_input["from_node"] = [0]
    line_input["to_node"] = [1]

    source_input = initialize_array(DatasetType.input, ComponentType.source, 1)
    source_input["id"] = [3]
    source_input["node"] = [0]
    source_input["status"] = [1]
    source_input["u_ref"] = [1.0]

    sym_voltage_sensor_input = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 1)
    sym_voltage_sensor_input["id"] = [4]
    sym_voltage_sensor_input["measured_object"] = [0]
    sym_voltage_sensor_input["u_sigma"] = [1.0]
    sym_voltage_sensor_input["u_measured"] = [0.0]

    sym_current_sensor_input = initialize_array(DatasetType.input, ComponentType.sym_current_sensor, 1)
    sym_current_sensor_input["id"] = [5]
    sym_current_sensor_input["measured_object"] = [2]
    sym_current_sensor_input["measured_terminal_type"] = [
        MeasuredTerminalType.branch_from,
    ]
    sym_current_sensor_input["angle_measurement_type"] = [
        AngleMeasurementType.global_angle,
    ]
    sym_current_sensor_input["i_sigma"] = [1.0]
    sym_current_sensor_input["i_measured"] = [0.0]

    model = PowerGridModel(
        input_data={
            ComponentType.node: node_input,
            ComponentType.line: line_input,
            ComponentType.source: source_input,
            ComponentType.sym_voltage_sensor: sym_voltage_sensor_input,
            ComponentType.sym_current_sensor: sym_current_sensor_input,
        }
    )

    with pytest.raises(NotObservableError):
        model._calculate_state_estimation(decode_error=True)


@pytest.mark.skip(reason="TODO")
def test_handle_power_grid_dataset_error():
    pass


@pytest.mark.skip(reason="TODO")
def test_handle_power_grid_unreachable_error():
    pass



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Type hints for the power-grid-model library.

This includes all miscellaneous type hints not under dataset or categories.
"""

from power_grid_model._core.dataset_definitions import (  # noqa: F401
    ComponentType,
    ComponentTypeVar,
    DatasetType,
    DatasetTypeVar,
)
from power_grid_model._core.power_grid_meta import (  # noqa: F401
    ComponentMetaData,
    DatasetMetaData,
    PowerGridMetaData,
)
from power_grid_model._core.typing import (
    ComponentAttributeMapping as _ComponentAttributeMapping,
)
from power_grid_model.enum import ComponentAttributeFilterOptions  # noqa: F401

ComponentAttributeMapping = _ComponentAttributeMapping
"""
Type hint for mapping component attributes.

`ComponentAttributeMapping` can be one of the following:

- A set of :class:`ComponentType` or `str`

- A list of :class:`ComponentType` or `str`

- A :class:`ComponentAttributeFilterOptions <power_grid_model.enum.ComponentAttributeFilterOptions>` value

- `None`

- A dictionary mapping :class:`ComponentType` to a set, list, `None` or
  :class:`ComponentAttributeFilterOptions <power_grid_model.enum.ComponentAttributeFilterOptions>`
"""



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd

from power_grid_model import ComponentType, DatasetType, LoadGenType, PowerGridModel, initialize_array

# node
node = initialize_array(DatasetType.input, ComponentType.node, 2)
# The following is also supported
# node = initialize_array("input", "node", 2)
node["id"] = [1, 2]
node["u_rated"] = [10.5e3, 10.5e3]

# line
line = initialize_array(DatasetType.input, ComponentType.line, 1)
line["id"] = [3]
line["from_node"] = [1]
line["to_node"] = [2]
line["from_status"] = [1]
line["to_status"] = [1]
line["r1"] = [0.25]
line["x1"] = [0.2]
line["c1"] = [10e-6]
line["tan1"] = [0.0]
line["i_n"] = [1000]

# load
sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
sym_load["id"] = [4]
sym_load["node"] = [2]
sym_load["status"] = [1]
sym_load["type"] = [LoadGenType.const_power]
sym_load["p_specified"] = [2e6]
sym_load["q_specified"] = [0.5e6]

# source
source = initialize_array(DatasetType.input, ComponentType.source, 1)
source["id"] = [5]
source["node"] = [1]
source["status"] = [1]
source["u_ref"] = [1.0]

# input_data
input_data = {
    ComponentType.node: node,
    ComponentType.line: line,
    ComponentType.sym_load: sym_load,
    ComponentType.source: source,
}

# call constructor
model = PowerGridModel(input_data, system_frequency=50.0)

result = model.calculate_power_flow()

print("Node Input")
print(pd.DataFrame(input_data[ComponentType.node]))
print("Node Result")
print(pd.DataFrame(result[ComponentType.node]))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Power grid model raw dataset handler
"""

from collections.abc import Mapping
from typing import Any, cast

from power_grid_model._core.buffer_handling import (
    BufferProperties,
    CAttributeBuffer,
    CBuffer,
    create_buffer,
    get_buffer_properties,
    get_buffer_view,
)
from power_grid_model._core.data_types import (
    AttributeType,
    ColumnarData,
    ComponentData,
    Dataset,
    DenseBatchColumnarData,
    SingleColumnarData,
    SparseBatchColumnarData,
)
from power_grid_model._core.dataset_definitions import ComponentType, DatasetType, _str_to_component_type
from power_grid_model._core.enum import ComponentAttributeFilterOptions
from power_grid_model._core.error_handling import VALIDATOR_MSG, assert_no_error
from power_grid_model._core.power_grid_core import (
    ConstDatasetPtr,
    DatasetInfoPtr,
    MutableDatasetPtr,
    WritableDatasetPtr,
    power_grid_core as pgc,
)
from power_grid_model._core.power_grid_meta import ComponentMetaData, DatasetMetaData, power_grid_meta_data
from power_grid_model._core.typing import ComponentAttributeMapping, _ComponentAttributeMappingDict
from power_grid_model._core.utils import (
    get_dataset_type,
    is_columnar,
    is_nan_or_equivalent,
    is_sparse,
    process_data_filter,
)


class CDatasetInfo:
    """
    Raw Power Grid Model dataset info.
    """

    def __init__(self, info: DatasetInfoPtr):
        self._info: DatasetInfoPtr = info

    def name(self) -> str:
        """
        The name of the dataset type.

        Returns:
            The name of the dataset type
        """
        return pgc.dataset_info_name(self._info)

    def dataset_type(self):
        """
        The name of the dataset type.

        Returns:
            The name of the dataset type
        """
        return self.name()

    def is_batch(self) -> bool:
        """
        Whether the dataset is a batch dataset.

        Returns:
            Whether the dataset is a batch dataset
        """
        return bool(pgc.dataset_info_is_batch(self._info))

    def batch_size(self) -> int:
        """
        The size of the dataset.

        Returns:
            The size of the dataset
        """
        return pgc.dataset_info_batch_size(self._info)

    def n_components(self) -> int:
        """
        The amount of components in the dataset.

        Returns:
            The amount of components in the dataset
        """
        return pgc.dataset_info_n_components(self._info)

    def components(self) -> list[ComponentType]:
        """
        The components in the dataset.

        Returns:
            A list of the component names in the dataset
        """
        return [
            _str_to_component_type(pgc.dataset_info_component_name(self._info, idx))
            for idx in range(self.n_components())
        ]

    def elements_per_scenario(self) -> Mapping[ComponentType, int]:
        """
        The number of elements per scenario in the dataset.

        Returns:
            The number of elements per senario for each component in the dataset;
            or -1 if the scenario is not uniform (different amount per scenario)
        """
        return {
            component_name: pgc.dataset_info_elements_per_scenario(self._info, idx)
            for idx, component_name in enumerate(self.components())
        }

    def total_elements(self) -> Mapping[ComponentType, int]:
        """
        The total number of elements in the dataset.

        Returns:
            The total number of elements for each component.
            For each component, if the number of elements per scenario is uniform, its value shall be equal to
                the product of the batch size and the amount of elements per scenario for that component.
        """
        return {
            component_name: pgc.dataset_info_total_elements(self._info, idx)
            for idx, component_name in enumerate(self.components())
        }

    def attribute_indications(self) -> Mapping[ComponentType, None | list[AttributeType]]:
        """
        The attribute indications in the dataset.

        Returns:
            A map of component to its attribute indications.
            None means no attribute indications
        """
        result_dict: dict[ComponentType, None | list[AttributeType]] = {}
        components = self.components()
        for component_idx, component_name in enumerate(components):
            has_indications = pgc.dataset_info_has_attribute_indications(self._info, component_idx)
            if has_indications == 0:
                result_dict[component_name] = None
            else:
                n_indications = pgc.dataset_info_n_attribute_indications(self._info, component_idx)
                result_dict[component_name] = [
                    pgc.dataset_info_attribute_name(self._info, component_idx, attribute_idx)
                    for attribute_idx in range(n_indications)
                ]
        return result_dict


class CMutableDataset:
    """
    A view of a user-owned dataset.

    This may be used to provide a user dataset to the Power Grid Model.

    The dataset will create mutable buffers that the Power Grid Model can use to load data.
    """

    _dataset_type: DatasetType
    _schema: DatasetMetaData
    _is_batch: bool
    _batch_size: int
    _mutable_dataset: MutableDatasetPtr
    _buffer_views: list[CBuffer]

    def __new__(cls, data: Dataset, dataset_type: Any = None):
        instance = super().__new__(cls)
        instance._mutable_dataset = MutableDatasetPtr()
        instance._buffer_views = []

        instance._dataset_type = dataset_type if dataset_type in DatasetType else get_dataset_type(data)
        instance._schema = power_grid_meta_data[instance._dataset_type]

        if data:
            first_component, first_component_data = next(iter(data.items()))
            first_sub_info = get_buffer_properties(data=first_component_data, schema=instance._schema[first_component])
            instance._is_batch = first_sub_info.is_batch
            instance._batch_size = first_sub_info.batch_size
        else:
            instance._is_batch = False
            instance._batch_size = 1

        instance._mutable_dataset = pgc.create_dataset_mutable(
            instance._dataset_type.value, instance._is_batch, instance._batch_size
        )
        assert_no_error()

        instance._add_data(data)
        assert_no_error()

        return instance

    def get_dataset_ptr(self) -> MutableDatasetPtr:
        """
        Get the raw underlying mutable dataset pointer.

        Returns:
            MutableDatasetPtr: the raw underlying mutable dataset pointer.
        """
        return self._mutable_dataset

    def get_info(self) -> CDatasetInfo:
        """
        Get the info for this dataset.

        Returns:
            The dataset info for this dataset.
        """
        return CDatasetInfo(pgc.dataset_mutable_get_info(self._mutable_dataset))

    def get_buffer_views(self) -> list[CBuffer]:
        """
        Get list of buffer views

        Returns:
            list of buffer view
        """
        return self._buffer_views

    def _add_data(self, data: Dataset):
        """
        Add Power Grid Model data to the mutable dataset view.

        Args:
            data: the data.

        Raises:
            ValueError: if the component is unknown.
            ValueError: if the data is inconsistent with the rest of the dataset.
            PowerGridError: if there was an internal error.
        """
        for component, component_data in data.items():
            self._add_component_data(component, component_data, allow_unknown=False)

    def _add_component_data(self, component: ComponentType, data: ComponentData, allow_unknown: bool = False):
        """
        Add Power Grid Model data for a single component to the mutable dataset view.

        Args:
            component: the name of the component
            data: the data of the component
            allow_unknown (optional): ignore any unknown components. Defaults to False.

        Raises:
            ValueError: if the component is unknown and allow_unknown is False.
            ValueError: if the data is inconsistent with the rest of the dataset.
            PowerGridError: if there was an internal error.
        """
        if component not in self._schema:
            if not allow_unknown:
                raise ValueError(f"Unknown component {component} in schema. {VALIDATOR_MSG}")
            return

        self._validate_properties(data, self._schema[component])
        c_buffer = get_buffer_view(data, self._schema[component], self._is_batch, self._batch_size)
        self._buffer_views.append(c_buffer)
        self._register_buffer(component, c_buffer)

    def _register_buffer(self, component: ComponentType, buffer: CBuffer):
        pgc.dataset_mutable_add_buffer(
            dataset=self._mutable_dataset,
            component=component.value,
            elements_per_scenario=buffer.n_elements_per_scenario,
            total_elements=buffer.total_elements,
            indptr=buffer.indptr,
            data=buffer.data,
        )
        assert_no_error()
        for attr, attr_data in buffer.attribute_data.items():
            self._register_attribute_buffer(component, attr, attr_data)

    def _register_attribute_buffer(self, component, attr, attr_data):
        pgc.dataset_mutable_add_attribute_buffer(
            dataset=self._mutable_dataset,
            component=component.value,
            attribute=attr,
            data=attr_data.data,
        )
        assert_no_error()

    def _validate_properties(self, data: ComponentData, schema: ComponentMetaData):
        properties = get_buffer_properties(data, schema=schema, is_batch=None, batch_size=None)
        if properties.is_batch != self._is_batch:
            raise ValueError(
                f"Dataset type (single or batch) must be consistent across all components. {VALIDATOR_MSG}"
            )
        if properties.batch_size != self._batch_size:
            raise ValueError(f"Dataset must have a consistent batch size across all components. {VALIDATOR_MSG}")

    def __del__(self):
        pgc.destroy_dataset_mutable(self._mutable_dataset)


class CConstDataset:
    """
    A view of a user-owned dataset.

    This may be used to provide a user dataset to the Power Grid Model.

    The dataset will create const buffers that the Power Grid Model can use to load data.

    It is created from mutable dataset.
    """

    _const_dataset: ConstDatasetPtr
    _buffer_views: list[CBuffer]

    def __new__(cls, data: Dataset, dataset_type: DatasetType | None = None):
        instance = super().__new__(cls)
        instance._const_dataset = ConstDatasetPtr()

        # create from mutable dataset
        mutable_dataset = CMutableDataset(data=data, dataset_type=dataset_type)
        instance._const_dataset = pgc.create_dataset_const_from_mutable(mutable_dataset.get_dataset_ptr())
        assert_no_error()
        instance._buffer_views = mutable_dataset.get_buffer_views()

        return instance

    def get_dataset_ptr(self) -> ConstDatasetPtr:
        """
        Get the raw underlying const dataset pointer.

        Returns:
            ConstDatasetPtr: the raw underlying const dataset pointer.
        """
        return self._const_dataset

    def get_info(self) -> CDatasetInfo:
        """
        Get the info for this dataset.

        Returns:
            The dataset info for this dataset.
        """
        return CDatasetInfo(pgc.dataset_const_get_info(self._const_dataset))

    def __del__(self):
        pgc.destroy_dataset_const(self._const_dataset)


class CWritableDataset:
    """
    A view of a Power Grid Model-owned dataset.

    This may be used to retrieve data from the Power Grid Model.

    This class provides buffers to which the Power Grid Model can write data in an external call.
    After writing to the buffers, the data contents can be retrieved.
    """

    def __init__(self, dataset_ptr: WritableDatasetPtr, data_filter: ComponentAttributeMapping):
        self._writable_dataset = dataset_ptr

        info = self.get_info()
        self._dataset_type = info.dataset_type()
        self._schema = power_grid_meta_data[self._dataset_type]

        self._data_filter = process_data_filter(
            dataset_type=info.dataset_type(),
            data_filter=data_filter,
            available_components=info.components(),
        )

        self._data: Dataset = {}
        self._buffers: Mapping[str, CBuffer] = {}
        self._component_buffer_properties = self._get_buffer_properties(info)

        self._add_buffers()
        assert_no_error()

    def get_dataset_ptr(self) -> WritableDatasetPtr:
        """
        Get the raw underlying writable dataset pointer.

        Returns:
            WritableDatasetPtr: the raw underlying writable dataset pointer.
        """
        return self._writable_dataset

    def get_info(self) -> CDatasetInfo:
        """
        Get the info for this dataset.

        Returns:
            The dataset info for this dataset.
        """
        return CDatasetInfo(pgc.dataset_writable_get_info(self._writable_dataset))

    def get_data(self) -> Dataset:
        """
        Retrieve data from the Power Grid Model dataset.

        The Power Grid Model may write to these buffers at a later point in time.

        Returns:
            The full dataset with filters applied.
        """
        self._post_filtering()
        return self._data

    def get_component_data(self, component: ComponentType) -> ComponentData:
        """
        Retrieve Power Grid Model data from the dataset for a specific component.

        Args:
            component: the component.

        Returns:
            The dataset for the specified component.
        """
        return self._data[component]

    def get_data_filter(self) -> _ComponentAttributeMappingDict:
        """Gets the data filter requested

        Returns:
            _ComponentAttributeMappingDict: data filter
        """
        return self._data_filter

    def _add_buffers(self):
        for component, buffer_properties in self._component_buffer_properties.items():
            self._add_buffer(component, buffer_properties)

    def _add_buffer(self, component: ComponentType, buffer_properties: BufferProperties):
        schema = self._schema[component]

        self._data[component] = create_buffer(buffer_properties, schema)
        self._register_buffer(component, get_buffer_view(self._data[component], schema))

    def _register_buffer(self, component: ComponentType, buffer: CBuffer):
        pgc.dataset_writable_set_buffer(
            dataset=self._writable_dataset,
            component=component,
            indptr=buffer.indptr,
            data=buffer.data,
        )
        assert_no_error()
        for attribute, attribute_data in buffer.attribute_data.items():
            self._register_attribute_buffer(component, attribute, attribute_data)

    def _register_attribute_buffer(
        self,
        component: ComponentType,
        attribute: AttributeType,
        buffer: CAttributeBuffer,
    ):
        pgc.dataset_writable_set_attribute_buffer(
            dataset=self._writable_dataset,
            component=component,
            attribute=attribute,
            data=buffer.data,
        )
        assert_no_error()

    def _get_buffer_properties(self, info: CDatasetInfo) -> Mapping[ComponentType, BufferProperties]:
        is_batch = info.is_batch()
        batch_size = info.batch_size()
        components = info.components()
        n_elements_per_scenario = info.elements_per_scenario()
        n_total_elements = info.total_elements()
        attribute_indications = info.attribute_indications()

        return {
            component: BufferProperties(
                is_sparse=n_elements_per_scenario[component] == -1,
                is_batch=is_batch,
                batch_size=batch_size,
                n_elements_per_scenario=n_elements_per_scenario[component],
                n_total_elements=n_total_elements[component],
                columns=_get_filtered_attributes(
                    schema=self._schema[component],
                    component_data_filter=self._data_filter[component],
                    attribute_indication=attribute_indications[component],
                ),
            )
            for component in components
            if component in self._data_filter
        }

    def _filter_attributes(self, buffer: ColumnarData):
        if is_sparse(buffer):
            attributes = cast(SparseBatchColumnarData, buffer)["data"]
        else:
            attributes = cast(SingleColumnarData | DenseBatchColumnarData, buffer)

        keys_to_remove = []
        for attr, array in attributes.items():
            if is_nan_or_equivalent(array):
                keys_to_remove.append(attr)
        for key in keys_to_remove:
            del attributes[key]

    def _filter_with_mapping(self):
        for component_type, component_buffer in self._data.items():
            if component_type in self._data_filter:
                filter_option = self._data_filter[component_type]
                if filter_option is ComponentAttributeFilterOptions.relevant and is_columnar(component_buffer):
                    self._filter_attributes(component_buffer)

    def _post_filtering(self):
        if isinstance(self._data_filter, dict):
            self._filter_with_mapping()


def _get_filtered_attributes(
    schema: ComponentMetaData,
    component_data_filter: set[str] | list[str] | None | ComponentAttributeFilterOptions,
    attribute_indication: None | list[AttributeType],
) -> list[AttributeType] | None:
    if component_data_filter is None:
        return None

    if isinstance(component_data_filter, ComponentAttributeFilterOptions):
        if component_data_filter == ComponentAttributeFilterOptions.relevant and attribute_indication is not None:
            return attribute_indication
        return [] if schema.dtype.names is None else list(schema.dtype.names)

    return list(component_data_filter)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import copy
from itertools import product
from typing import cast
from unittest.mock import ANY, MagicMock, patch

import numpy as np
import pytest

from power_grid_model import CalculationType, LoadGenType, MeasuredTerminalType, initialize_array, power_grid_meta_data
from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.enum import AngleMeasurementType
from power_grid_model._core.utils import compatibility_convert_row_columnar_dataset
from power_grid_model.enum import Branch3Side, BranchSide, ComponentAttributeFilterOptions, FaultType
from power_grid_model.validation import assert_valid_input_data
from power_grid_model.validation._validation import (
    assert_valid_data_structure,
    validate_generic_current_sensor,
    validate_generic_power_sensor,
    validate_ids,
    validate_input_data,
    validate_required_values,
    validate_unique_ids_across_components,
    validate_values,
)
from power_grid_model.validation.errors import (
    IdNotInDatasetError,
    InfinityError,
    InvalidAssociatedEnumValueError,
    InvalidEnumValueError,
    InvalidIdError,
    MissingValueError,
    MultiComponentNotUniqueError,
    NotUniqueError,
    PQSigmaPairError,
    UnsupportedMeasuredTerminalType,
)

NaN = cast(int, power_grid_meta_data[DatasetType.input][ComponentType.node].nans["id"])


def test_assert_valid_data_structure():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 3)
    line_input = initialize_array(DatasetType.input, ComponentType.line, 3)
    node_update = initialize_array(DatasetType.update, ComponentType.node, 3)
    line_update = initialize_array(DatasetType.update, ComponentType.line, 3)

    # Input data: Assertion ok
    assert_valid_data_structure({ComponentType.node: node_input, ComponentType.line: line_input}, DatasetType.input)

    # Update data: Assertion ok
    assert_valid_data_structure({ComponentType.node: node_update, ComponentType.line: line_update}, DatasetType.update)

    # There is no such thing as 'output' data
    with pytest.raises(KeyError, match=r"output"):
        assert_valid_data_structure({ComponentType.node: node_input, ComponentType.line: line_input}, "output")

    # Input data is not valid update data
    with pytest.raises(TypeError, match=r"Unexpected Numpy structured array"):
        assert_valid_data_structure(
            {ComponentType.node: node_input, ComponentType.line: line_input}, DatasetType.update
        )

    # Update data is not valid input data
    with pytest.raises(TypeError, match=r"Unexpected Numpy structured array"):
        assert_valid_data_structure(
            {ComponentType.node: node_update, ComponentType.line: line_update}, DatasetType.input
        )

    # A normal numpy array is not valid input data
    node_dummy = np.array([(1, 10.5e3), (2, 10.5e3), (3, 10.5e3)])
    with pytest.raises(TypeError, match=r"Unexpected Numpy array"):
        assert_valid_data_structure({ComponentType.node: node_dummy, ComponentType.line: line_input}, DatasetType.input)

    # A structured numpy array, with wrong data type for u_rated f4 != f8, is not valid input data
    node_dummy = np.array([(1, 10.5e3), (2, 10.5e3), (3, 10.5e3)], dtype=[("id", "i4"), ("u_rated", "f4")])
    with pytest.raises(TypeError, match=r"Unexpected Numpy structured array"):
        assert_valid_data_structure({ComponentType.node: node_dummy, ComponentType.line: line_input}, DatasetType.input)

    # A structured numpy array, with correct data types is not a valid, is still not valid input data.
    # N.B. It is not 'aligned'
    node_dummy = np.array([(1, 10.5e3), (2, 10.5e3), (3, 10.5e3)], dtype=[("id", "i4"), ("u_rated", "f8")])
    with pytest.raises(TypeError, match=r"Unexpected Numpy structured array"):
        assert_valid_data_structure({ComponentType.node: node_dummy, ComponentType.line: line_input}, DatasetType.input)

    # Invalid component type
    input_with_wrong_component = {ComponentType.node: node_input, "some_random_component": line_input}
    with pytest.raises(KeyError, match=r"Unknown component \'some_random_component\' in DatasetType\.input data"):
        assert_valid_data_structure(input_with_wrong_component, DatasetType.input)

    input_with_wrong_data_type = {ComponentType.node: node_input, ComponentType.line: [1, 2, 3]}
    with pytest.raises(
        TypeError, match=r"Unexpected data type list for \'ComponentType\.line\' DatasetType\.input data"
    ):
        assert_valid_data_structure(input_with_wrong_data_type, DatasetType.input)


def test_validate_unique_ids_across_components():
    node = initialize_array(DatasetType.input, ComponentType.node, 3)
    node["id"] = [1, 2, 3]

    line = initialize_array(DatasetType.input, ComponentType.line, 3)
    line["id"] = [4, 5, 3]

    transformer = initialize_array(DatasetType.input, ComponentType.transformer, 3)
    transformer["id"] = [1, 6, 7]

    source = initialize_array(DatasetType.input, ComponentType.source, 3)
    source["id"] = [8, 9, 10]

    input_data = {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.transformer: transformer,
        ComponentType.source: source,
    }

    unique_id_errors = validate_unique_ids_across_components(input_data)

    assert (
        MultiComponentNotUniqueError(
            [(ComponentType.node, "id"), (ComponentType.line, "id"), (ComponentType.transformer, "id")],
            [(ComponentType.node, 1), (ComponentType.node, 3), (ComponentType.line, 3), (ComponentType.transformer, 1)],
        )
        in unique_id_errors
    )
    n_non_unique_ids = 4
    assert len(unique_id_errors[0].ids) == n_non_unique_ids


def test_validate_ids():
    source = initialize_array(DatasetType.input, ComponentType.source, 3)
    source["id"] = [1, 2, 3]

    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 3)
    sym_load["id"] = [4, 5, 6]

    input_data = {
        ComponentType.source: source,
        ComponentType.sym_load: sym_load,
    }

    source_update = initialize_array(DatasetType.update, ComponentType.source, 3)
    source_update["id"] = [1, 2, 4]
    source_update["u_ref"] = [1.0, 2.0, 3.0]

    sym_load_update = initialize_array(DatasetType.update, ComponentType.sym_load, 3)
    sym_load_update["id"] = [4, 5, 7]
    sym_load_update["p_specified"] = [4.0, 5.0, 6.0]

    update_data = {ComponentType.source: source_update, ComponentType.sym_load: sym_load_update}

    invalid_ids = validate_ids(update_data, input_data)

    assert IdNotInDatasetError(ComponentType.source, [4], DatasetType.update) in invalid_ids
    assert IdNotInDatasetError(ComponentType.sym_load, [7], DatasetType.update) in invalid_ids

    source_update_no_id = initialize_array(DatasetType.update, ComponentType.source, 3)
    source_update_no_id["u_ref"] = [1.0, 2.0, 3.0]

    update_data_col = compatibility_convert_row_columnar_dataset(
        data={ComponentType.source: source_update_no_id, ComponentType.sym_load: sym_load_update},
        data_filter=ComponentAttributeFilterOptions.relevant,
        dataset_type=DatasetType.update,
    )
    invalid_ids = validate_ids(update_data_col, input_data)
    assert len(invalid_ids) == 1
    assert IdNotInDatasetError(ComponentType.sym_load, [7], DatasetType.update) in invalid_ids

    source_update_less_no_id = initialize_array(DatasetType.update, ComponentType.source, 2)
    source_update_less_no_id["u_ref"] = [1.0, 2.0]

    update_data_col_less_no_id = compatibility_convert_row_columnar_dataset(
        data={ComponentType.source: source_update_less_no_id, ComponentType.sym_load: sym_load_update},
        data_filter=ComponentAttributeFilterOptions.relevant,
        dataset_type=DatasetType.update,
    )
    invalid_ids = validate_ids(update_data_col_less_no_id, input_data)
    n_invalid_ids_update_source_without_id = 2
    assert len(invalid_ids) == n_invalid_ids_update_source_without_id
    assert IdNotInDatasetError(ComponentType.sym_load, [7], DatasetType.update) in invalid_ids

    source_update_part_nan_id = initialize_array(DatasetType.update, ComponentType.source, 3)
    source_update_part_nan_id["id"] = [1, np.iinfo(np.int32).min, 4]
    source_update_part_nan_id["u_ref"] = [1.0, 2.0, 3.0]

    update_data_col_part_nan_id = compatibility_convert_row_columnar_dataset(
        data={ComponentType.source: source_update_part_nan_id, ComponentType.sym_load: sym_load_update},
        data_filter=ComponentAttributeFilterOptions.relevant,
        dataset_type=DatasetType.update,
    )
    invalid_ids = validate_ids(update_data_col_part_nan_id, input_data)
    n_invalid_ids_source_update = 2
    assert len(invalid_ids) == n_invalid_ids_source_update
    assert IdNotInDatasetError(ComponentType.sym_load, [7], DatasetType.update) in invalid_ids


@pytest.mark.parametrize(
    "calculation_type",
    [
        pytest.param(None, id="no calculation type specified"),
        pytest.param(CalculationType.power_flow, id="power_flow"),
        pytest.param(CalculationType.state_estimation, id="state_estimation"),
        pytest.param(CalculationType.short_circuit, id="short_circuit"),
    ],
)
@pytest.mark.parametrize("symmetric", [pytest.param(True, id="symmetric"), pytest.param(False, id="asymmetric")])
def test_validate_required_values_sym_calculation(calculation_type, symmetric):
    node = initialize_array(DatasetType.input, ComponentType.node, 1)
    line = initialize_array(DatasetType.input, ComponentType.line, 1)
    link = initialize_array(DatasetType.input, ComponentType.link, 1)
    transformer = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    three_winding_transformer = initialize_array(DatasetType.input, ComponentType.three_winding_transformer, 1)
    source = initialize_array(DatasetType.input, ComponentType.source, 1)
    shunt = initialize_array(DatasetType.input, ComponentType.shunt, 1)
    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
    sym_gen = initialize_array(DatasetType.input, ComponentType.sym_gen, 1)
    asym_load = initialize_array(DatasetType.input, ComponentType.asym_load, 1)
    asym_gen = initialize_array(DatasetType.input, ComponentType.asym_gen, 1)
    sym_voltage_sensor = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 1)

    asym_voltage_sensor = initialize_array(DatasetType.input, ComponentType.asym_voltage_sensor, 1)
    asym_voltage_sensor["u_measured"] = [[1.0, np.nan, 2.0]]

    sym_power_sensor = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 1)

    asym_power_sensor = initialize_array(DatasetType.input, ComponentType.asym_power_sensor, 1)
    asym_power_sensor["p_measured"] = [[np.nan, 2.0, 1.0]]
    asym_power_sensor["q_measured"] = [[2.0, 1.0, np.nan]]

    sym_current_sensor = initialize_array(DatasetType.input, ComponentType.sym_current_sensor, 1)

    asym_current_sensor = initialize_array(DatasetType.input, ComponentType.asym_current_sensor, 1)
    asym_current_sensor["i_measured"] = [[np.nan, 2.0, 1.0]]
    asym_current_sensor["i_angle_measured"] = [[2.0, 1.0, np.nan]]

    fault = initialize_array(DatasetType.input, ComponentType.fault, 1)

    data = {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.link: link,
        ComponentType.transformer: transformer,
        ComponentType.three_winding_transformer: three_winding_transformer,
        ComponentType.source: source,
        ComponentType.shunt: shunt,
        ComponentType.sym_load: sym_load,
        ComponentType.sym_gen: sym_gen,
        ComponentType.asym_load: asym_load,
        ComponentType.asym_gen: asym_gen,
        ComponentType.sym_voltage_sensor: sym_voltage_sensor,
        ComponentType.asym_voltage_sensor: asym_voltage_sensor,
        ComponentType.sym_power_sensor: sym_power_sensor,
        ComponentType.asym_power_sensor: asym_power_sensor,
        ComponentType.sym_current_sensor: sym_current_sensor,
        ComponentType.asym_current_sensor: asym_current_sensor,
        ComponentType.fault: fault,
    }
    required_values_errors = validate_required_values(data=data, calculation_type=calculation_type, symmetric=symmetric)

    pf_dependent = calculation_type == CalculationType.power_flow or calculation_type is None
    se_dependent = calculation_type == CalculationType.state_estimation or calculation_type is None
    sc_dependent = calculation_type == CalculationType.short_circuit or calculation_type is None
    asym_dependent = not symmetric

    assert MissingValueError(ComponentType.node, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.node, "u_rated", [NaN]) in required_values_errors

    assert MissingValueError(ComponentType.line, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "from_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "to_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "from_status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "to_status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "r1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "x1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "c1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "tan1", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.line, "r0", [NaN]) in required_values_errors) == asym_dependent
    assert (MissingValueError(ComponentType.line, "x0", [NaN]) in required_values_errors) == asym_dependent
    assert (MissingValueError(ComponentType.line, "c0", [NaN]) in required_values_errors) == asym_dependent
    assert (MissingValueError(ComponentType.line, "tan0", [NaN]) in required_values_errors) == asym_dependent

    # i_n made optional later in lines
    assert MissingValueError(ComponentType.line, "i_n", [NaN]) not in required_values_errors

    assert MissingValueError(ComponentType.link, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.link, "from_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.link, "to_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.link, "from_status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.link, "to_status", [NaN]) in required_values_errors

    assert MissingValueError(ComponentType.transformer, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "from_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "to_node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "from_status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "to_status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "u1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "u2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "sn", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "uk", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "pk", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "i0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "p0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "winding_from", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "winding_to", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "clock", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "tap_side", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "tap_min", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "tap_max", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.transformer, "tap_size", [NaN]) in required_values_errors

    assert MissingValueError(ComponentType.three_winding_transformer, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "node_1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "node_2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "node_3", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "status_1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "status_2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "status_3", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "u1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "u2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "u3", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "sn_1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "sn_2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "sn_3", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "uk_12", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "uk_13", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "uk_23", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "pk_12", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "pk_13", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "pk_23", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "i0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "p0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "winding_1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "winding_2", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "winding_3", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "clock_12", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "clock_13", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "tap_side", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "tap_min", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "tap_max", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.three_winding_transformer, "tap_size", [NaN]) in required_values_errors

    assert MissingValueError(ComponentType.source, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.source, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.source, "status", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.source, "u_ref", [NaN]) in required_values_errors) == pf_dependent

    assert MissingValueError(ComponentType.shunt, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.shunt, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.shunt, "status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.shunt, "g1", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.shunt, "b1", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.shunt, "g0", [NaN]) in required_values_errors) == asym_dependent
    assert (MissingValueError(ComponentType.shunt, "b0", [NaN]) in required_values_errors) == asym_dependent

    assert MissingValueError(ComponentType.sym_load, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_load, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_load, "status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_load, "type", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.sym_load, "p_specified", [NaN]) in required_values_errors) == pf_dependent
    assert (MissingValueError(ComponentType.sym_load, "q_specified", [NaN]) in required_values_errors) == pf_dependent

    assert MissingValueError(ComponentType.sym_gen, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_gen, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_gen, "status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_gen, "type", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.sym_gen, "p_specified", [NaN]) in required_values_errors) == pf_dependent
    assert (MissingValueError(ComponentType.sym_gen, "q_specified", [NaN]) in required_values_errors) == pf_dependent

    assert MissingValueError(ComponentType.asym_load, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_load, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_load, "status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_load, "type", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.asym_load, "p_specified", [NaN]) in required_values_errors) == pf_dependent
    assert (MissingValueError(ComponentType.asym_load, "q_specified", [NaN]) in required_values_errors) == pf_dependent

    assert MissingValueError(ComponentType.asym_gen, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_gen, "node", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_gen, "status", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_gen, "type", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.asym_gen, "p_specified", [NaN]) in required_values_errors) == pf_dependent
    assert (MissingValueError(ComponentType.asym_gen, "q_specified", [NaN]) in required_values_errors) == pf_dependent

    assert MissingValueError(ComponentType.sym_voltage_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_voltage_sensor, "measured_object", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.sym_voltage_sensor, "u_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_voltage_sensor, "u_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.asym_voltage_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_voltage_sensor, "measured_object", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.asym_voltage_sensor, "u_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_voltage_sensor, "u_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.sym_power_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_power_sensor, "measured_object", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_power_sensor, "measured_terminal_type", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.sym_power_sensor, "power_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_power_sensor, "p_measured", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_power_sensor, "q_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.asym_power_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_power_sensor, "measured_object", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_power_sensor, "measured_terminal_type", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.asym_power_sensor, "power_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_power_sensor, "p_measured", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_power_sensor, "q_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.sym_current_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.sym_current_sensor, "measured_object", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "measured_terminal_type", [NaN]) in required_values_errors
    )
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "angle_measurement_type", [NaN]) in required_values_errors
    )
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "i_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "i_angle_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "i_measured", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.sym_current_sensor, "i_angle_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.asym_current_sensor, "id", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.asym_current_sensor, "measured_object", [NaN]) in required_values_errors
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "measured_terminal_type", [NaN]) in required_values_errors
    )
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "angle_measurement_type", [NaN]) in required_values_errors
    )
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "i_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "i_angle_sigma", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "i_measured", [NaN]) in required_values_errors
    ) == se_dependent
    assert (
        MissingValueError(ComponentType.asym_current_sensor, "i_angle_measured", [NaN]) in required_values_errors
    ) == se_dependent

    assert MissingValueError(ComponentType.fault, "id", [NaN]) in required_values_errors
    assert (MissingValueError(ComponentType.fault, "status", [NaN]) in required_values_errors) == sc_dependent
    assert (MissingValueError(ComponentType.fault, "fault_type", [NaN]) in required_values_errors) == sc_dependent


def test_validate_required_values_asym_calculation():
    line = initialize_array(DatasetType.input, ComponentType.line, 1)
    shunt = initialize_array(DatasetType.input, ComponentType.shunt, 1)

    data = {ComponentType.line: line, ComponentType.shunt: shunt}
    required_values_errors = validate_required_values(data=data, symmetric=False)

    assert MissingValueError(ComponentType.line, "r0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "x0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "c0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.line, "tan0", [NaN]) in required_values_errors

    assert MissingValueError(ComponentType.shunt, "g0", [NaN]) in required_values_errors
    assert MissingValueError(ComponentType.shunt, "b0", [NaN]) in required_values_errors


@pytest.mark.parametrize("fault_types", product(list(FaultType), list(FaultType)))
def test_validate_fault_sc_calculation(fault_types):
    line = initialize_array(DatasetType.input, ComponentType.line, 1)
    shunt = initialize_array(DatasetType.input, ComponentType.shunt, 1)
    fault = initialize_array(DatasetType.input, ComponentType.fault, 2)
    fault["fault_type"] = fault_types

    data = {ComponentType.line: line, ComponentType.shunt: shunt, ComponentType.fault: fault}
    required_values_errors = validate_required_values(data=data, calculation_type=CalculationType.short_circuit)

    asym_sc_calculation = np.any(
        list(fault_type not in (FaultType.three_phase, FaultType.nan) for fault_type in fault_types)
    )

    assert (MissingValueError(ComponentType.line, "r0", [NaN]) in required_values_errors) == asym_sc_calculation
    assert (MissingValueError(ComponentType.line, "x0", [NaN]) in required_values_errors) == asym_sc_calculation
    assert (MissingValueError(ComponentType.line, "c0", [NaN]) in required_values_errors) == asym_sc_calculation
    assert (MissingValueError(ComponentType.line, "tan0", [NaN]) in required_values_errors) == asym_sc_calculation

    assert (MissingValueError(ComponentType.shunt, "g0", [NaN]) in required_values_errors) == asym_sc_calculation
    assert (MissingValueError(ComponentType.shunt, "b0", [NaN]) in required_values_errors) == asym_sc_calculation


def test_validate_values():
    # Create invalid nodes and lines
    node = initialize_array(DatasetType.input, ComponentType.node, 3)
    line = initialize_array(DatasetType.input, ComponentType.line, 3)

    # Validate nodes and lines individually
    node_errors = validate_values({ComponentType.node: node})
    line_errors = validate_values({ComponentType.line: line})

    # Validate nodes and lines combined
    both_errors = validate_values({ComponentType.node: node, ComponentType.line: line})

    # The errors should add up (in this simple case)
    assert both_errors == node_errors + line_errors


def test_validate_values__calculation_types():
    # Create invalid sensor
    sym_voltage_sensor = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 3)
    all_errors = validate_values({ComponentType.sym_voltage_sensor: sym_voltage_sensor})
    power_flow_errors = validate_values(
        {ComponentType.sym_voltage_sensor: sym_voltage_sensor}, calculation_type=CalculationType.power_flow
    )
    state_estimation_errors = validate_values(
        {ComponentType.sym_voltage_sensor: sym_voltage_sensor}, calculation_type=CalculationType.state_estimation
    )

    assert not power_flow_errors
    assert all_errors == state_estimation_errors


@pytest.mark.parametrize(
    ("sensor_type", "parameter"),
    [
        (ComponentType.sym_voltage_sensor, "u_sigma"),
        (ComponentType.asym_voltage_sensor, "u_sigma"),
        (ComponentType.sym_power_sensor, "power_sigma"),
        (ComponentType.sym_power_sensor, "p_sigma"),
        (ComponentType.sym_power_sensor, "q_sigma"),
        (ComponentType.asym_power_sensor, "power_sigma"),
        (ComponentType.asym_power_sensor, "p_sigma"),
        (ComponentType.asym_power_sensor, "q_sigma"),
        (ComponentType.sym_current_sensor, "i_sigma"),
        (ComponentType.sym_current_sensor, "i_angle_sigma"),
        (ComponentType.asym_current_sensor, "i_sigma"),
        (ComponentType.asym_current_sensor, "i_angle_sigma"),
    ],
)
def test_validate_values__infinite_sigmas(sensor_type, parameter):
    sensor_array = initialize_array(DatasetType.input, sensor_type, 3)
    sensor_array[parameter] = np.inf
    all_errors = validate_values({sensor_type: sensor_array})

    for error in all_errors:
        assert not isinstance(error, InfinityError)


@pytest.mark.parametrize(
    ("sensor_type", "values", "error_types"),
    [
        pytest.param(
            ComponentType.sym_power_sensor,
            [[np.nan, np.nan], [], []],
            [InvalidIdError, NotUniqueError],
            id="both nan - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[np.inf, np.inf], [], []],
            [InvalidIdError, NotUniqueError],
            id="both inf - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[0.1, np.nan], [], []],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="q_sigma nan - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[0.1, np.inf], [], []],
            [InvalidIdError, NotUniqueError],
            id="q_sigma inf - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[np.nan, 0.1], [], []],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="p_sigma nan - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[np.inf, 0.1], [], []],
            [InvalidIdError, NotUniqueError],
            id="p_sigma inf - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.sym_power_sensor,
            [[0.1, 0.1], [], []],
            [InvalidIdError, NotUniqueError],
            id="no nan or inf - sym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[np.nan, np.nan, np.nan]] * 3, [[np.nan, np.nan, np.nan]] * 3],
            [InvalidIdError, NotUniqueError],
            id="all nan - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[np.inf, np.inf, np.inf]] * 3, [[np.inf, np.inf, np.inf]] * 3],
            [InvalidIdError, NotUniqueError],
            id="all inf - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, np.nan, 0.1], [0.1, np.nan, 0.1], [0.1, 0.1, 0.1]],
                [[np.nan, 0.1, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="mixed nans for some sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, np.inf, 0.1], [0.1, np.inf, 0.1], [0.1, 0.1, 0.1]],
                [[np.inf, 0.1, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="mixed infs for some sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[0.1, np.nan, np.nan]] * 3, [[np.nan, np.nan, np.nan]] * 3],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="all nan except one phase p_sigma - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[0.1, np.inf, np.inf]] * 3, [[np.inf, np.inf, np.inf]] * 3],
            [InvalidIdError, NotUniqueError],
            id="all inf except one phase p_sigma - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[np.nan, np.nan, np.nan]] * 3, [[0.1, np.nan, np.nan]] * 3],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="all nan except one phase q_sigma - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[np.inf, np.inf, np.inf]] * 3, [[0.1, np.inf, np.inf]] * 3],
            [InvalidIdError, NotUniqueError],
            id="all inf except one phase q_sigma - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, 0.1, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[np.nan, np.nan, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="one sensor mixed nans - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, 0.1, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[np.inf, np.inf, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="one sensor mixed infs - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[np.nan, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="q_sigma of one sensor nan - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[np.inf, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="q_sigma of one sensor inf - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[np.nan, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="p_sigma of one sensor nan - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[np.inf, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="p_sigma of one sensor inf - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[np.nan, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="all nan except one attribute for one phase for one sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[np.inf, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="all inf except one attribute for one phase for one sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, np.nan, np.nan], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError, PQSigmaPairError],
            id="all nan except one phase for one sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [
                [],
                [[0.1, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
                [[0.1, np.inf, np.inf], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]],
            ],
            [InvalidIdError, NotUniqueError],
            id="all inf except one phase for one sensor - asym_power_sensor",
        ),
        pytest.param(
            ComponentType.asym_power_sensor,
            [[], [[0.1, 0.1, 0.1]] * 3, [[0.1, 0.1, 0.1]] * 3],
            [InvalidIdError, NotUniqueError],
            id="no nan or inf - asym_power_sensor",
        ),
    ],
)
def test_validate_values__bad_p_q_sigma(sensor_type, values, error_types):
    def arbitrary_fill(array, sensor_type, values):
        if sensor_type == ComponentType.sym_power_sensor:
            array["p_sigma"] = values[0][0]
            array["q_sigma"] = values[0][1]
        else:
            array["p_sigma"][0] = values[1][0]
            array["p_sigma"][1] = values[1][1]
            array["p_sigma"][2] = values[1][2]
            array["q_sigma"][0] = values[2][0]
            array["q_sigma"][1] = values[2][1]
            array["q_sigma"][2] = values[2][2]
        array["id"] = [123, 234, 345]

    sensor_array = initialize_array(DatasetType.input, sensor_type, 3)
    arbitrary_fill(sensor_array, sensor_type, values)
    all_errors = validate_values({sensor_type: sensor_array})

    for error in all_errors:
        assert any(isinstance(error, error_type) for error_type in error_types)
        assert set(error.ids).issubset(set(sensor_array["id"]))


@pytest.mark.parametrize(
    ("values", "error_types"),
    [
        ([[np.nan, np.nan], [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]], [InvalidIdError]),
        (
            [[0.1, np.nan], [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]],
            [InvalidIdError, PQSigmaPairError],
        ),
        (
            [[np.nan, np.nan], [[np.nan, 0.1, np.nan], [np.nan, np.nan, np.nan]]],
            [InvalidIdError, PQSigmaPairError],
        ),
        (
            [[np.nan, np.nan], [[np.nan, 0.1, np.nan], [np.nan, 0.1, np.nan]]],
            [InvalidIdError, PQSigmaPairError],
        ),
        (
            [[0.1, 0.1], [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]],
            [InvalidIdError, PQSigmaPairError],
        ),
        ([[0.1, 0.1], [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]], [InvalidIdError]),
        ([[np.nan, np.nan], [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1]]], [InvalidIdError]),
        ([[0.1, 0.1], [[0.1, 0.1, 0.1], [0.1, 0.1, 0.1]]], [InvalidIdError]),
    ],
)
def test_validate_values__bad_p_q_sigma_both_components(values, error_types):
    def two_component_data(values):
        node = initialize_array(DatasetType.input, ComponentType.node, 1)
        node["id"] = 123
        sym_power_sensor = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 1)
        sym_power_sensor["p_sigma"] = values[0][0]
        sym_power_sensor["q_sigma"] = values[0][1]
        sym_power_sensor["id"] = 456
        asym_power_sensor = initialize_array(DatasetType.input, ComponentType.asym_power_sensor, 1)
        asym_power_sensor["p_measured"] = values[1][0]
        asym_power_sensor["q_measured"] = values[1][1]
        asym_power_sensor["id"] = 789

        return {
            ComponentType.node: node,
            ComponentType.sym_power_sensor: sym_power_sensor,
            ComponentType.asym_power_sensor: asym_power_sensor,
        }

    data = two_component_data(values)
    all_errors = validate_values(data)
    for error in all_errors:
        assert any(isinstance(error, error_type) for error_type in error_types)
        assert (data[error.component]["id"] == error.ids).all()


def test_validate_values__bad_p_q_sigma_single_component_twice():
    def single_component_twice_data():
        node = initialize_array(DatasetType.input, ComponentType.node, 1)
        node["id"] = 123
        sym_power_sensor = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 2)
        sym_power_sensor["p_sigma"] = [np.nan, 0.1]
        sym_power_sensor["q_sigma"] = [np.nan, np.nan]
        sym_power_sensor["id"] = [456, 789]

        return {
            ComponentType.node: node,
            ComponentType.sym_power_sensor: sym_power_sensor,
        }

    data = single_component_twice_data()
    all_errors = validate_values(data)
    for error in all_errors:
        assert any(isinstance(error, error_type) for error_type in [InvalidIdError, PQSigmaPairError])
        if isinstance(error, PQSigmaPairError):
            assert error.ids[0] == data["sym_power_sensor"]["id"][1]


@pytest.mark.parametrize("measured_terminal_type", MeasuredTerminalType)
@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_ids")
def test_validate_generic_power_sensor__all_terminal_types(
    _all_valid_ids: MagicMock, measured_terminal_type: MeasuredTerminalType
):
    # Act
    validate_generic_power_sensor(data={}, component="")  # type: ignore

    # Assert
    _all_valid_ids.assert_any_call(
        ANY, ANY, field=ANY, ref_components=ANY, measured_terminal_type=measured_terminal_type
    )


@pytest.mark.parametrize(
    ("ref_component", "measured_terminal_type"),
    [
        (
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            MeasuredTerminalType.branch_from,
        ),
        (
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            MeasuredTerminalType.branch_to,
        ),
        (ComponentType.source, MeasuredTerminalType.source),
        (ComponentType.shunt, MeasuredTerminalType.shunt),
        ([ComponentType.sym_load, ComponentType.asym_load], MeasuredTerminalType.load),
        ([ComponentType.sym_gen, ComponentType.asym_gen], MeasuredTerminalType.generator),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_1),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_2),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_3),
        (ComponentType.node, MeasuredTerminalType.node),
    ],
)
@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_ids")
def test_validate_generic_power_sensor__terminal_types(
    _all_valid_ids: MagicMock,
    ref_component: ComponentType | list[ComponentType],
    measured_terminal_type: MeasuredTerminalType,
):
    # Act
    validate_generic_power_sensor(data={}, component="")  # type: ignore

    # Assert
    _all_valid_ids.assert_any_call(
        ANY, ANY, field=ANY, ref_components=ref_component, measured_terminal_type=measured_terminal_type
    )


@pytest.mark.parametrize(
    "measured_terminal_type",
    [
        MeasuredTerminalType.branch_from,
        MeasuredTerminalType.branch_to,
        MeasuredTerminalType.branch3_1,
        MeasuredTerminalType.branch3_2,
        MeasuredTerminalType.branch3_3,
    ],
)
@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_in_valid_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal", new=MagicMock())
@patch(
    "power_grid_model.validation._validation._any_voltage_angle_measurement_if_global_current_measurement",
    new=MagicMock(),
)
@patch("power_grid_model.validation._validation._all_valid_ids")
def test_validate_generic_current_sensor__all_terminal_types(
    _all_valid_ids: MagicMock, measured_terminal_type: MeasuredTerminalType
):
    # Act
    validate_generic_current_sensor(data={}, component="")  # type: ignore

    # Assert
    _all_valid_ids.assert_any_call(
        ANY,
        ANY,
        field=ANY,
        ref_components=ANY,
        measured_terminal_type=measured_terminal_type,
    )


@pytest.mark.parametrize("current_sensor_type", [ComponentType.sym_current_sensor, ComponentType.asym_current_sensor])
@pytest.mark.parametrize(
    ("measured_terminal_type", "supported"),
    [
        (MeasuredTerminalType.branch_from, True),
        (MeasuredTerminalType.branch_to, True),
        (MeasuredTerminalType.branch3_1, True),
        (MeasuredTerminalType.branch3_2, True),
        (MeasuredTerminalType.branch3_3, True),
        (MeasuredTerminalType.source, False),
        (MeasuredTerminalType.shunt, False),
        (MeasuredTerminalType.load, False),
        (MeasuredTerminalType.generator, False),
        (MeasuredTerminalType.node, False),
    ],
)
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock(return_value=[]))
@patch("power_grid_model.validation._validation._all_valid_ids", new=MagicMock(return_value=[]))
@patch(
    "power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal",
    new=MagicMock(return_value=[]),
)
@patch(
    "power_grid_model.validation._validation._any_voltage_angle_measurement_if_global_current_measurement",
    new=MagicMock(return_value=[]),
)
@patch(
    "power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal",
    new=MagicMock(return_value=[]),
)
def test_validate_generic_current_sensor__only_branches_supported(
    current_sensor_type: ComponentType, measured_terminal_type: MeasuredTerminalType, supported: bool
):
    current_sensor_data = initialize_array(DatasetType.input, current_sensor_type, 1)
    current_sensor_data["id"] = 1
    current_sensor_data["measured_terminal_type"] = measured_terminal_type

    result = validate_generic_current_sensor(
        data={current_sensor_type: current_sensor_data}, component=current_sensor_type
    )

    if supported:
        assert not result
    else:
        assert result == [
            UnsupportedMeasuredTerminalType(
                current_sensor_type,
                "measured_terminal_type",
                [1],
                [
                    MeasuredTerminalType.branch_from,
                    MeasuredTerminalType.branch_to,
                    MeasuredTerminalType.branch3_1,
                    MeasuredTerminalType.branch3_2,
                    MeasuredTerminalType.branch3_3,
                ],
            )
        ]


@pytest.mark.parametrize(
    ("ref_component", "measured_terminal_type"),
    [
        (
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            MeasuredTerminalType.branch_from,
        ),
        (
            [ComponentType.line, ComponentType.asym_line, ComponentType.generic_branch, ComponentType.transformer],
            MeasuredTerminalType.branch_to,
        ),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_1),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_2),
        (ComponentType.three_winding_transformer, MeasuredTerminalType.branch3_3),
    ],
)
@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_in_valid_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal", new=MagicMock())
@patch(
    "power_grid_model.validation._validation._any_voltage_angle_measurement_if_global_current_measurement",
    new=MagicMock(),
)
@patch("power_grid_model.validation._validation._all_valid_ids")
def test_validate_generic_current_sensor__terminal_types(
    _all_valid_ids: MagicMock,
    ref_component: ComponentType | list[ComponentType],
    measured_terminal_type: MeasuredTerminalType,
):
    # Act
    validate_generic_current_sensor(data={}, component="")  # type: ignore

    # Assert
    _all_valid_ids.assert_any_call(
        ANY,
        ANY,
        field=ANY,
        ref_components=ref_component,
        measured_terminal_type=measured_terminal_type,
    )


@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_in_valid_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_ids", new=MagicMock())
@patch(
    "power_grid_model.validation._validation._any_voltage_angle_measurement_if_global_current_measurement",
    new=MagicMock(),
)
@patch("power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal")
def test_validate_generic_current_sensor__angle_measurement_type_mixing(
    _all_same_current_angle_measurement_type_on_terminal,
):
    # Act
    validate_generic_current_sensor(data={}, component="")  # type: ignore

    # Assert
    _all_same_current_angle_measurement_type_on_terminal.assert_any_call(
        ANY,
        ANY,
        measured_object_field="measured_object",
        measured_terminal_type_field="measured_terminal_type",
        angle_measurement_type_field="angle_measurement_type",
    )


@patch("power_grid_model.validation._validation.validate_base", new=MagicMock())
@patch("power_grid_model.validation._validation._all_greater_than_zero", new=MagicMock())
@patch("power_grid_model.validation._validation._all_in_valid_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_enum_values", new=MagicMock())
@patch("power_grid_model.validation._validation._all_valid_ids", new=MagicMock())
@patch("power_grid_model.validation._validation._all_same_current_angle_measurement_type_on_terminal", new=MagicMock())
@patch("power_grid_model.validation._validation._any_voltage_angle_measurement_if_global_current_measurement")
def test_any_voltage_angle_measurement_if_global_current_measurement(
    any_voltage_angle_measurement_if_global_current_measurement,
):
    # Act
    validate_generic_current_sensor(data={}, component="")

    # Assert
    any_voltage_angle_measurement_if_global_current_measurement.assert_any_call(
        ANY,
        ANY,
        angle_measurement_type_filter=("angle_measurement_type", AngleMeasurementType.global_angle),
        voltage_sensor_u_angle_measured={
            ComponentType.sym_voltage_sensor: "u_angle_measured",
            ComponentType.asym_voltage_sensor: "u_angle_measured",
        },
    )


@pytest.mark.parametrize("power_sensor_type", [ComponentType.sym_power_sensor, ComponentType.asym_power_sensor])
@pytest.mark.parametrize("current_sensor_type", [ComponentType.sym_current_sensor, ComponentType.asym_current_sensor])
@patch("power_grid_model.validation._validation._all_finite", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_node", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_line", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_asym_line", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_branch", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_branch", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_transformer", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_three_winding_transformer", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_source", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_load_gen", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_load_gen", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_load_gen", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_load_gen", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_shunt", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_voltage_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_voltage_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_power_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_power_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_current_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_generic_current_sensor", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_fault", new=MagicMock())
@patch("power_grid_model.validation._validation.validate_transformer_tap_regulator", new=MagicMock())
@patch("power_grid_model.validation._validation._all_same_sensor_type_on_same_terminal")
def test_no_power_and_current_sensor_on_same_terminal(
    _all_same_sensor_type_on_same_terminal,
    power_sensor_type: ComponentType,
    current_sensor_type: ComponentType,
):
    # Act
    validate_values({power_sensor_type: None, current_sensor_type: None})

    # Assert
    _all_same_sensor_type_on_same_terminal.assert_any_call(
        ANY,
        power_sensor_type=power_sensor_type,
        current_sensor_type=current_sensor_type,
        measured_object_field="measured_object",
        measured_terminal_type_field="measured_terminal_type",
    )


def test_power_sigma_or_p_q_sigma():
    # node
    node = initialize_array(DatasetType.input, ComponentType.node, 2)
    node["id"] = np.array([0, 3])
    node["u_rated"] = [10.5e3, 10.5e3]

    # line
    line = initialize_array(DatasetType.input, ComponentType.line, 1)
    line["id"] = [2]
    line["from_node"] = [0]
    line["to_node"] = [3]
    line["from_status"] = [1]
    line["to_status"] = [1]
    line["r1"] = [0.001]
    line["x1"] = [0.02]
    line["c1"] = [0.0]
    line["tan1"] = [0.0]
    line["i_n"] = [1000.0]

    # load
    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 2)
    sym_load["id"] = [4, 9]
    sym_load["node"] = [3, 0]
    sym_load["status"] = [1, 1]
    sym_load["type"] = [LoadGenType.const_power, LoadGenType.const_power]
    sym_load["p_specified"] = [1e6, 1e6]
    sym_load["q_specified"] = [-1e6, -1e6]

    # source
    source = initialize_array(DatasetType.input, ComponentType.source, 1)
    source["id"] = [1]
    source["node"] = [0]
    source["status"] = [1]
    source["u_ref"] = [1.0]

    # voltage sensor
    voltage_sensor = initialize_array(DatasetType.input, ComponentType.sym_voltage_sensor, 1)
    voltage_sensor["id"] = 5
    voltage_sensor["measured_object"] = 0
    voltage_sensor["u_sigma"] = [100.0]
    voltage_sensor["u_measured"] = [10.5e3]

    # power sensor
    sym_power_sensor = initialize_array(DatasetType.input, ComponentType.sym_power_sensor, 3)
    sym_power_sensor["id"] = [6, 7, 8]
    sym_power_sensor["measured_object"] = [2, 4, 9]
    sym_power_sensor["measured_terminal_type"] = [
        MeasuredTerminalType.branch_from,
        MeasuredTerminalType.load,
        MeasuredTerminalType.load,
    ]
    sym_power_sensor["p_measured"] = [1e6, -1e6, -1e6]
    sym_power_sensor["q_measured"] = [1e6, -1e6, -1e6]
    sym_power_sensor["power_sigma"] = [np.nan, 1e9, 1e9]
    sym_power_sensor["p_sigma"] = [1e4, np.nan, 1e4]
    sym_power_sensor["q_sigma"] = [1e9, np.nan, 1e9]

    # power sensor
    asym_power_sensor = initialize_array(DatasetType.input, ComponentType.asym_power_sensor, 4)
    asym_power_sensor["id"] = [66, 77, 88, 99]
    asym_power_sensor["measured_object"] = [2, 4, 9, 9]
    asym_power_sensor["measured_terminal_type"] = [
        MeasuredTerminalType.branch_from,
        MeasuredTerminalType.load,
        MeasuredTerminalType.load,
        MeasuredTerminalType.load,
    ]
    asym_power_sensor["p_measured"] = [[1e6, 1e6, 1e6], [-1e6, -1e6, -1e6], [-1e6, -1e6, -1e6], [-1e6, -1e6, -1e6]]
    asym_power_sensor["q_measured"] = [[1e6, 1e6, 1e6], [-1e6, -1e6, -1e6], [-1e6, -1e6, -1e6], [-1e6, -1e6, -1e6]]
    asym_power_sensor["power_sigma"] = [np.nan, 1e9, 1e9, 1e9]
    asym_power_sensor["p_sigma"] = [[1e4, 1e4, 1e4], [np.nan, np.nan, np.nan], [1e4, 1e4, 1e4], [1e4, 1e4, 1e4]]
    asym_power_sensor["q_sigma"] = [[1e9, 1e9, 1e9], [np.nan, np.nan, np.nan], [1e9, 1e4, 1e4], [1e9, 1e4, 1e4]]

    # all
    input_data = {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.sym_load: sym_load,
        ComponentType.source: source,
        ComponentType.sym_voltage_sensor: voltage_sensor,
        ComponentType.sym_power_sensor: sym_power_sensor,
        ComponentType.asym_power_sensor: asym_power_sensor,
    }

    assert_valid_input_data(input_data=input_data, calculation_type=CalculationType.state_estimation)

    np.testing.assert_array_equal(sym_power_sensor["power_sigma"], [np.nan, 1e9, 1e9])
    np.testing.assert_array_equal(sym_power_sensor["p_sigma"], [1e4, np.nan, 1e4])
    np.testing.assert_array_equal(sym_power_sensor["q_sigma"], [1e9, np.nan, 1e9])
    np.testing.assert_array_equal(asym_power_sensor["power_sigma"], [np.nan, 1e9, 1e9, 1e9])
    np.testing.assert_array_equal(
        asym_power_sensor["p_sigma"], [[1e4, 1e4, 1e4], [np.nan, np.nan, np.nan], [1e4, 1e4, 1e4], [1e4, 1e4, 1e4]]
    )
    np.testing.assert_array_equal(
        asym_power_sensor["q_sigma"], [[1e9, 1e9, 1e9], [np.nan, np.nan, np.nan], [1e9, 1e4, 1e4], [1e9, 1e4, 1e4]]
    )

    # bad weather
    bad_input_data = copy.deepcopy(input_data)
    bad_sym_power_sensor = bad_input_data[ComponentType.sym_power_sensor]
    bad_sym_power_sensor["power_sigma"] = [np.nan, np.nan, 1e9]
    bad_sym_power_sensor["p_sigma"] = [np.nan, np.nan, 1e4]
    bad_sym_power_sensor["q_sigma"] = [np.nan, 1e9, np.nan]
    errors = validate_input_data(input_data=bad_input_data, calculation_type=CalculationType.state_estimation)
    n_sym_input_validation_errors = 4
    assert len(errors) == n_sym_input_validation_errors
    assert errors == [
        MissingValueError(ComponentType.sym_power_sensor, "power_sigma", [6]),
        MissingValueError(ComponentType.sym_power_sensor, "p_sigma", [7]),
        MissingValueError(ComponentType.sym_power_sensor, "q_sigma", [8]),
        PQSigmaPairError(ComponentType.sym_power_sensor, ("p_sigma", "q_sigma"), [7, 8]),
    ]

    np.testing.assert_array_equal(bad_sym_power_sensor["power_sigma"], [np.nan, np.nan, 1e9])
    np.testing.assert_array_equal(bad_sym_power_sensor["p_sigma"], [np.nan, np.nan, 1e4])
    np.testing.assert_array_equal(bad_sym_power_sensor["q_sigma"], [np.nan, 1e9, np.nan])

    # bad weather
    bad_input_data = copy.deepcopy(input_data)
    bad_asym_power_sensor = bad_input_data[ComponentType.asym_power_sensor]
    bad_asym_power_sensor["power_sigma"] = [np.nan, np.nan, 1e9, np.nan]
    bad_asym_power_sensor["p_sigma"] = [
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
        [1e4, np.nan, np.nan],
        [1e4, np.nan, np.nan],
    ]
    bad_asym_power_sensor["q_sigma"] = [
        [np.nan, np.nan, np.nan],
        [1e9, 1e9, 1e9],
        [np.nan, 1e4, 1e4],
        [np.nan, 1e4, 1e4],
    ]
    errors = validate_input_data(input_data=bad_input_data, calculation_type=CalculationType.state_estimation)
    n_asym_input_validation_errors = 4
    assert len(errors) == n_asym_input_validation_errors
    assert errors == [
        MissingValueError(ComponentType.asym_power_sensor, "power_sigma", [66]),
        MissingValueError(ComponentType.asym_power_sensor, "p_sigma", [77, 88, 99]),
        MissingValueError(ComponentType.asym_power_sensor, "q_sigma", [88, 99]),
        PQSigmaPairError(ComponentType.asym_power_sensor, ("p_sigma", "q_sigma"), [77, 88, 99]),
    ]

    np.testing.assert_array_equal(bad_asym_power_sensor["power_sigma"], [np.nan, np.nan, 1e9, np.nan])
    np.testing.assert_array_equal(
        bad_asym_power_sensor["p_sigma"],
        [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan], [1e4, np.nan, np.nan], [1e4, np.nan, np.nan]],
    )
    np.testing.assert_array_equal(
        bad_asym_power_sensor["q_sigma"],
        [[np.nan, np.nan, np.nan], [1e9, 1e9, 1e9], [np.nan, 1e4, 1e4], [np.nan, 1e4, 1e4]],
    )


def test_all_default_values():
    """
    Initialize all components that have attributes that have default values, without setting values for
    those attributes.
    """
    node = initialize_array(DatasetType.input, ComponentType.node, 3)
    node["id"] = [0, 1, 2]
    node["u_rated"] = [50.0e3, 20.0e3, 10.5e3]

    source = initialize_array(DatasetType.input, ComponentType.source, 1)
    source["id"] = [3]
    source["node"] = [2]
    source["status"] = [1]
    source["u_ref"] = [1.0]

    transformer = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    transformer["id"] = [4]
    transformer["from_node"] = [0]
    transformer["to_node"] = [2]
    transformer["from_status"] = [1]
    transformer["to_status"] = [1]
    transformer["u1"] = [50e3]
    transformer["u2"] = [10.5e3]
    transformer["sn"] = [1e5]
    transformer["uk"] = [0.1]
    transformer["pk"] = [1e3]
    transformer["i0"] = [1.0e-6]
    transformer["p0"] = [0.1]
    transformer["winding_from"] = [2]
    transformer["winding_to"] = [1]
    transformer["clock"] = [5]
    transformer["tap_side"] = [0]
    transformer["tap_min"] = [-11]
    transformer["tap_max"] = [9]
    transformer["tap_size"] = [100]

    three_winding_transformer = initialize_array(DatasetType.input, ComponentType.three_winding_transformer, 1)
    three_winding_transformer["id"] = [6]
    three_winding_transformer["node_1"] = [0]
    three_winding_transformer["node_2"] = [1]
    three_winding_transformer["node_3"] = [2]
    three_winding_transformer["status_1"] = [1]
    three_winding_transformer["status_2"] = [1]
    three_winding_transformer["status_3"] = [1]
    three_winding_transformer["u1"] = [50.0e3]
    three_winding_transformer["u2"] = [20.0e3]
    three_winding_transformer["u3"] = [10.5e3]
    three_winding_transformer["sn_1"] = [1e5]
    three_winding_transformer["sn_2"] = [1e5]
    three_winding_transformer["sn_3"] = [1e5]
    three_winding_transformer["uk_12"] = [0.09]
    three_winding_transformer["uk_13"] = [0.06]
    three_winding_transformer["uk_23"] = [0.06]
    three_winding_transformer["pk_12"] = [1e3]
    three_winding_transformer["pk_13"] = [1e3]
    three_winding_transformer["pk_23"] = [1e3]
    three_winding_transformer["i0"] = [0]
    three_winding_transformer["p0"] = [0]
    three_winding_transformer["winding_1"] = [2]
    three_winding_transformer["winding_2"] = [1]
    three_winding_transformer["winding_3"] = [1]
    three_winding_transformer["clock_12"] = [5]
    three_winding_transformer["clock_13"] = [-7]  # supports periodic clock input
    three_winding_transformer["tap_side"] = [0]
    three_winding_transformer["tap_min"] = [-10]
    three_winding_transformer["tap_max"] = [10]
    three_winding_transformer["tap_size"] = [1380]

    fault = initialize_array(DatasetType.input, ComponentType.fault, 1)
    fault["id"] = [5]
    fault["status"] = [1]
    fault["fault_object"] = [0]

    input_data = {
        ComponentType.node: node,
        ComponentType.transformer: transformer,
        ComponentType.three_winding_transformer: three_winding_transformer,
        ComponentType.source: source,
        ComponentType.fault: fault,
    }

    assert_valid_input_data(input_data=input_data, calculation_type=CalculationType.power_flow)


@patch("power_grid_model.validation._validation.validate_transformer", new=MagicMock(return_value=[]))
@patch("power_grid_model.validation._validation.validate_three_winding_transformer", new=MagicMock(return_value=[]))
def test_validate_values__tap_regulator_control_side():
    # Create valid transformer
    transformer = initialize_array(DatasetType.input, ComponentType.transformer, 4)
    transformer["id"] = [0, 1, 2, 3]
    transformer["tap_side"] = [BranchSide.from_side, BranchSide.from_side, BranchSide.from_side, BranchSide.from_side]

    # Create valid three winding transformer
    three_winding_transformer = initialize_array(DatasetType.input, ComponentType.three_winding_transformer, 3)
    three_winding_transformer["id"] = [4, 5, 6]
    three_winding_transformer["tap_side"] = [Branch3Side.side_1, Branch3Side.side_1, Branch3Side.side_1]

    # Create invalid regulator
    transformer_tap_regulator = initialize_array(DatasetType.input, ComponentType.transformer_tap_regulator, 7)
    transformer_tap_regulator["id"] = np.arange(7, 14)
    transformer_tap_regulator["status"] = 1
    transformer_tap_regulator["regulated_object"] = np.arange(7)
    transformer_tap_regulator["control_side"] = [
        BranchSide.to_side,  # OK
        BranchSide.from_side,  # OK
        Branch3Side.side_3,  # branch3 provided but it is a 2-winding transformer (invalid)
        10,  # control side entirely out of range (invalid)
        Branch3Side.side_3,  # OK
        Branch3Side.side_1,  # OK
        10,  # control side entirely out of range (invalid)
    ]

    input_data = {
        ComponentType.transformer: transformer,
        ComponentType.three_winding_transformer: three_winding_transformer,
        ComponentType.transformer_tap_regulator: transformer_tap_regulator,
    }
    all_errors = validate_values(input_data)
    power_flow_errors = validate_values(input_data, calculation_type=CalculationType.power_flow)
    state_estimation_errors = validate_values(input_data, calculation_type=CalculationType.state_estimation)

    assert power_flow_errors == all_errors
    assert not state_estimation_errors

    n_input_validation_errors = 3

    assert len(all_errors) == n_input_validation_errors
    assert (
        InvalidEnumValueError(
            ComponentType.transformer_tap_regulator, "control_side", [10, 13], [BranchSide, Branch3Side]
        )
        in all_errors
    )
    assert (
        InvalidAssociatedEnumValueError(
            ComponentType.transformer_tap_regulator,
            ["control_side", "regulated_object"],
            [9, 10],
            [BranchSide],
        )
        in all_errors
    )
    assert (
        InvalidAssociatedEnumValueError(
            ComponentType.transformer_tap_regulator,
            ["control_side", "regulated_object"],
            [13],
            [Branch3Side],
        )
        in all_errors
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from copy import deepcopy
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from meta_data import AllDatasetMapData, AttributeClass, DatasetMapData, DatasetMetaData

DATA_DIR = Path(__file__).parent / "data"
TEMPLATE_DIR = Path(__file__).parent / "templates"
OUTPUT_PATH = Path(__file__).parents[1]

JINJA_ENV = Environment(loader=FileSystemLoader(TEMPLATE_DIR), autoescape=True)


def _data_type_nan(data_type: str):
    if data_type == "ID":
        return "na_IntID"
    if data_type == "double" or "RealValue" in data_type:
        return "nan"
    if data_type == "IntS":
        return "na_IntS"
    return f"static_cast<{data_type}>(na_IntS)"


class CodeGenerator:
    all_classes: dict[str, AttributeClass]
    base_output_path: Path

    def __init__(self, base_output_path: Path) -> None:
        self.all_classes = {}
        self.base_output_path = base_output_path

    def render_template(self, template_path: Path, output_path: Path, **data):
        # jinja expects a string, representing a relative path with forward slashes
        template_path_str = str(template_path.relative_to(TEMPLATE_DIR)).replace("\\", "/")
        template = JINJA_ENV.get_template(template_path_str)

        output = template.render(**data)

        with output_path.open(mode="w", encoding="utf-8") as output_file:
            output_file.write(output)

    def render_attribute_classes(self, template_path: Path, data_path: Path, output_path: Path):
        with data_path.open() as data_file:
            json_data = data_file.read()
        dataset_meta_data: DatasetMetaData = DatasetMetaData.schema().loads(json_data)
        # flatten attribute list
        # assign full name
        for attribute_class in dataset_meta_data.classes:
            if attribute_class.is_template:
                attribute_class.full_name = f"{attribute_class.name}<sym>"
                attribute_class.specification_names = [
                    f"{attribute_class.name}<symmetric_t>",
                    f"{attribute_class.name}<asymmetric_t>",
                    f"Sym{attribute_class.name}",
                    f"Asym{attribute_class.name}",
                ]
            else:
                attribute_class.full_name = attribute_class.name
                attribute_class.specification_names = [attribute_class.name]
            new_attribute_list = []
            for attribute in attribute_class.attributes:
                attribute.nan_value = _data_type_nan(attribute.data_type)
                if isinstance(attribute.names, str):
                    new_attribute_list.append(attribute)
                else:
                    # flatten list
                    for name in attribute.names:
                        new_attribute = deepcopy(attribute)
                        new_attribute.names = name
                        new_attribute_list.append(new_attribute)
            attribute_class.attributes = new_attribute_list
            # get full attribute
            if attribute_class.base is not None:
                base_class = next(filter(lambda x: x.name == attribute_class.base, dataset_meta_data.classes))
                attribute_class.full_attributes = base_class.full_attributes + attribute_class.attributes
                attribute_class.base_attributes = base_class.base_attributes.copy()
                attribute_class.base_attributes[base_class.name] = base_class.full_attributes
            else:
                attribute_class.full_attributes = attribute_class.attributes
                attribute_class.base_attributes = {}
            # add to class dict
            self.all_classes[attribute_class.name] = attribute_class

        self.render_template(
            template_path=template_path,
            output_path=output_path,
            classes=dataset_meta_data.classes,
            include_guard=dataset_meta_data.include_guard,
            name=dataset_meta_data.name,
        )

    def render_dataset_class_maps(self, template_path: Path, data_path: Path, output_path: Path):
        with data_path.open() as data_file:
            json_data = data_file.read()
        dataset_meta_data: list[DatasetMapData] = AllDatasetMapData.schema().loads(json_data).all_datasets

        # create list
        all_map = {}
        for dataset in dataset_meta_data:
            prefixes = ["sym_", "asym_"] if dataset.is_template else [""]
            for prefix in prefixes:
                all_components = {}
                for component in dataset.components:
                    class_def: AttributeClass = self.all_classes[component.class_name]
                    for component_name in component.names:
                        all_components[component_name] = [x.names for x in class_def.full_attributes]
                all_map[f"{prefix}{dataset.name}"] = all_components
        self.render_template(template_path=template_path, output_path=output_path, all_map=all_map)

    def code_gen(self):
        render_funcs = {
            "attribute_classes": self.render_attribute_classes,
            "dataset_class_maps": self.render_dataset_class_maps,
        }

        # render attribute classes
        for template_name, render_func in render_funcs.items():
            for template_path in TEMPLATE_DIR.rglob(f"{template_name}.*.jinja"):
                output_suffix = template_path.with_suffix("").suffix
                output_dir = template_path.parent.relative_to(TEMPLATE_DIR)
                for data_path in DATA_DIR.glob(f"{template_name}/*.json"):
                    output_path = self.base_output_path / output_dir / data_path.with_suffix(output_suffix).name
                    output_path.parent.mkdir(parents=True, exist_ok=True)
                    print(f"Generating file: {output_path}")
                    render_func(template_path=template_path, data_path=data_path, output_path=output_path)


if __name__ == "__main__":
    CodeGenerator(OUTPUT_PATH).code_gen()



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from enum import IntEnum

import pytest

from power_grid_model import ComponentType
from power_grid_model.validation.errors import (
    ComparisonError,
    InvalidAssociatedEnumValueError,
    InvalidEnumValueError,
    InvalidIdError,
    MultiComponentValidationError,
    MultiFieldValidationError,
    SingleFieldValidationError,
    ValidationError,
)


def test_validation_error():
    error = ValidationError()
    assert str(error) == "An unknown validation error occurred."
    assert error.component_str == str(None)
    assert error.field_str == str(None)
    assert not error.ids


def test_single_field_validation_error():
    error = SingleFieldValidationError(component=ComponentType.node, field="bravo", ids=[0, 1, 1, 2, 3, 5])
    assert error.component_str == "node"
    assert error.field_str == "'bravo'"
    assert str(error) == "Field 'bravo' is not valid for 6 nodes."


def test_multi_field_validation_error():
    error = MultiFieldValidationError(component=ComponentType.node, fields=["delta", "echo"], ids=[0, 1, 1, 2, 3, 5])
    assert error.component_str == "node"
    assert error.field_str == "'delta' and 'echo'"
    assert str(error) == "Combination of fields 'delta' and 'echo' is not valid for 6 nodes."

    with pytest.raises(ValueError, match="at least 2 fields"):
        MultiFieldValidationError(component=ComponentType.node, fields=["delta"], ids=[])


def test_multi_component_validation_error():
    error = MultiComponentValidationError(
        fields=[(ComponentType.node, "golf"), (ComponentType.line, "india")],
        ids=[
            (ComponentType.node, 0),
            (ComponentType.node, 1),
            (ComponentType.node, 1),
            (ComponentType.line, 2),
            (ComponentType.line, 3),
            (ComponentType.line, 5),
        ],
    )
    assert error.component_str == "line/node"
    assert error.field_str == "line.india and node.golf"
    assert str(error) == "Fields line.india and node.golf are not valid for 6 lines/nodes."

    with pytest.raises(ValueError, match="at least 2 fields"):
        MultiComponentValidationError(fields=[(ComponentType.node, "golf")], ids=[])

    with pytest.raises(ValueError, match="at least 2 components"):
        MultiComponentValidationError(fields=[(ComponentType.node, "golf"), (ComponentType.node, "india")], ids=[])


def test_invalid_enum_value_error():
    class CustomType(IntEnum):
        pass

    error = InvalidEnumValueError(component=ComponentType.node, field="lima", ids=[1, 2, 3], enum=CustomType)
    assert error.component == ComponentType.node
    assert error.field == "lima"
    assert error.ids == [1, 2, 3]
    assert error.enum is CustomType
    assert str(error) == "Field 'lima' contains invalid CustomType values for 3 nodes."


def test_invalid_id_error():
    error = InvalidIdError(ComponentType.node, field="november", ids=[1, 2, 3], ref_components=["oscar", "papa"])
    assert error.component == ComponentType.node
    assert error.field == "november"
    assert error.ids == [1, 2, 3]
    assert error.ref_components == ["oscar", "papa"]
    assert str(error) == "Field 'november' does not contain a valid oscar/papa id for 3 nodes."


def test_invalid_id_error_with_filters():
    error = InvalidIdError(
        ComponentType.node,
        field="november",
        ids=[1, 2, 3],
        ref_components=["oscar", "papa"],
        filters={"foo": "bar", "baz": ComponentType.node},
    )
    assert error.component == ComponentType.node
    assert error.field == "november"
    assert error.ids == [1, 2, 3]
    assert error.ref_components == ["oscar", "papa"]
    assert error.filters_str == "(foo=bar, baz=node)"
    assert str(error) == "Field 'november' does not contain a valid oscar/papa id for 3 nodes. (foo=bar, baz=node)"


def test_comparison_error():
    error = ComparisonError(component=ComponentType.node, field="romeo", ids=[1, 2, 3], ref_value=0)
    assert error.component == ComponentType.node
    assert error.field == "romeo"
    assert error.ids == [1, 2, 3]
    assert error.ref_value == 0
    assert ComparisonError(component="", field="", ids=[], ref_value=123).ref_value_str == "123"
    assert ComparisonError(component="", field="", ids=[], ref_value=123.456).ref_value_str == "123.456"
    assert ComparisonError(component="", field="", ids=[], ref_value=0).ref_value_str == "zero"
    assert ComparisonError(component="", field="", ids=[], ref_value=0.0).ref_value_str == "zero"
    assert ComparisonError(component="", field="", ids=[], ref_value=1).ref_value_str == "one"
    assert ComparisonError(component="", field="", ids=[], ref_value=1.0).ref_value_str == "one"
    assert ComparisonError(component="", field="", ids=[], ref_value="sierra").ref_value_str == "sierra"


def test_error_context():
    error = ComparisonError(component=ComponentType.node, field="tango", ids=[1, 2, 3], ref_value=0)
    context = error.get_context()
    expected_context_keys = 4
    assert len(context) == expected_context_keys
    assert context["component"] == ComponentType.node
    assert context["field"] == "'tango'"
    assert context["ids"] == [1, 2, 3]
    assert context["ref_value"] == "zero"

    context = error.get_context(id_lookup=["Uniform", "Victor"])  # Note that id 0 does not exist in the error ids
    assert context["ids"] == {1: "Victor", 2: None, 3: None}

    context = error.get_context(id_lookup={1: "Victor", 3: "Whiskey"})
    assert context["ids"] == {1: "Victor", 2: None, 3: "Whiskey"}


def test_error_context_tuple_ids():
    error = MultiComponentValidationError(
        fields=[(ComponentType.node, "x"), ("b", "y")], ids=[(ComponentType.node, 1), ("b", 2), (ComponentType.node, 3)]
    )
    context = error.get_context(id_lookup={1: "Victor", 3: "Whiskey"})
    assert context["ids"] == {("node", 1): "Victor", ("b", 2): None, ("node", 3): "Whiskey"}


def test_invalid_associated_enum_value_error():
    class CustomType(IntEnum):
        pass

    error = InvalidAssociatedEnumValueError(
        component=ComponentType.node, fields=["bar", "baz"], ids=[1, 2], enum=[CustomType]
    )
    assert error.component == ComponentType.node
    assert error.field == ["bar", "baz"]
    assert error.ids == [1, 2]
    assert len(error.enum) == 1
    for actual, expected in zip(error.enum, [CustomType]):
        assert actual is expected
    assert str(error) == "The combination of fields 'bar' and 'baz' results in invalid CustomType values for 2 nodes."



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Data types involving PGM datasets.

Data types for library-internal use. In an attempt to clarify type hints, some types
have been defined and explained in this file.
"""

from typing import TypeAlias, TypedDict, TypeVar

import numpy as np

from power_grid_model._core.dataset_definitions import ComponentTypeVar

SingleArray: TypeAlias = np.ndarray

AttributeType: TypeAlias = str

SingleColumn: TypeAlias = np.ndarray

DenseBatchArray: TypeAlias = np.ndarray

SingleColumnarData = dict[AttributeType, SingleColumn]

_SingleComponentData = TypeVar("_SingleComponentData", SingleArray, SingleColumnarData)  # deduction helper
SingleComponentData = SingleArray | SingleColumnarData


SingleDataset = dict[ComponentTypeVar, _SingleComponentData]

BatchList = list[SingleDataset]

BatchColumn: TypeAlias = np.ndarray

DenseBatchColumnarData = dict[AttributeType, BatchColumn]

IndexPointer: TypeAlias = np.ndarray


class SparseBatchColumnarData(TypedDict):
    """
    Sparse batch columnar data is a dictionary containing the keys `indptr` and `data`.

    - data: a :class:`SingleColumnarData`. The exact supported attribute columns depend on the component type.
    - indptr: an :class:`IndexPointer` representing the start and end indices for each batch scenario.

    - Examples:

        - structure: {"indptr": :class:`IndexPointer`, "data": :class:`SingleColumnarData`}
        - concrete example: {"indptr": [0, 2, 2, 3], "data": {"id": [0, 1, 0], "status": [1, 1, 0]}}

            - the scenario 0 sets the status of components with ids 0 and 1 to 1
              (and keeps defaults for other components)
            - scenario 1 keeps the default values for all components
            - scenario 2 sets the status of component with id 0 to 0 (and keeps defaults for other components)
    """

    indptr: IndexPointer
    data: SingleColumnarData


class SparseBatchArray(TypedDict):
    """
    A sparse batch array is a dictionary containing the keys `indptr` and `data`.

    - data: a :class:`SingleArray`. The exact dtype depends on the type of component.
    - indptr: an :class:`IndexPointer` representing the start and end indices for each batch scenario.

    - Examples:

        - structure: {"indptr": :class:`IndexPointer`, "data": :class:`SingleArray`}
        - concrete example: {"indptr": [0, 2, 2, 3], "data": [(0, 1, 1), (1, 1, 1), (0, 0, 0)]}

            - the scenario 0 sets the statuses of components with ids 0 and 1 to 1
              (and keeps defaults for other components)
            - scenario 1 keeps the default values for all components
            - scenario 2 sets the statuses of component with id 0 to 0 (and keeps defaults for other components)
    """

    indptr: IndexPointer
    data: SingleArray


SparseBatchData = SparseBatchArray | SparseBatchColumnarData

SparseDataComponentType: TypeAlias = str

BatchColumnarData = DenseBatchColumnarData | SparseBatchColumnarData

ColumnarData = SingleColumnarData | BatchColumnarData
BatchArray = DenseBatchArray | SparseBatchArray


BatchComponentData = BatchArray | BatchColumnarData

_BatchComponentData = TypeVar("_BatchComponentData", BatchArray, BatchColumnarData)  # deduction helper


BatchDataset = dict[ComponentTypeVar, _BatchComponentData]


DataArray = SingleArray | BatchArray


_ComponentData = TypeVar("_ComponentData", SingleComponentData, BatchComponentData)  # deduction helper
ComponentData = DataArray | ColumnarData

Dataset = dict[ComponentTypeVar, _ComponentData]


DenseBatchData = DenseBatchArray | DenseBatchColumnarData

NominalValue = int

RealValue = float

AsymValue = tuple[RealValue, RealValue, RealValue]

AttributeValue = RealValue | NominalValue | AsymValue

Component = dict[AttributeType, AttributeValue | str]

ComponentList = list[Component]

SinglePythonDataset = dict[ComponentTypeVar, ComponentList]

BatchPythonDataset = list[SinglePythonDataset]

PythonDataset = SinglePythonDataset | BatchPythonDataset
